#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#


from threading import Event, Timer


class RepeatingTimer(Timer):
    """A timer timing out several times instead of just once.

    Note that the next time is called after the function has finished!

    :param float interval: Interval between readouts in s.
    """

    def __init__(self, interval, function, args=None, kwargs=None):
        super().__init__(interval, function, args, kwargs)
        self.daemon = True

    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


class SignallingTimer(RepeatingTimer):
    """Repeating timer that sets an Event (:attr:`signal`) at timeout and continues counting.

    :param float interval: Interval in s.
    """

    def __init__(self, interval):
        self.signal = Event()
        super().__init__(interval, self._timeout, args=(self.signal,))

    @staticmethod
    def _timeout(signal):
        """Set and clear the signal event."""
        signal.set()
        signal.clear()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import warnings

# Maximally flexible approach to obtain version numbers, based on this approach:
# https://github.com/pypa/setuptools_scm/issues/143#issuecomment-672878863
# Sadly, this does not work with editable installs, which bake in version info on installation.
# see also https://github.com/pyusb/pyusb/pull/307#issuecomment-650797688
try:
    # If a user has setuptools_scm installed, assume they want the most up to date version string.
    # Alternatively, we could use a dummy dev module that is never packaged whose presence signals
    # that we are in an editable install/repo, see https://github.com/pycalphad/pycalphad/pull/341
    import setuptools_scm  # type: ignore
    __version__ = setuptools_scm.get_version(root='..', relative_to=__file__)
    del setuptools_scm
except (ImportError, LookupError):  # pragma: no cover
    # Setuptools_scm was not found, or it could not find a version, so use installation metadata.
    from importlib.metadata import version, PackageNotFoundError
    try:
        __version__ = version("pyleco")
        # Alternatively, if the current approach is too slow, we could add
        # 'write_to = "pyleco/_version.py"' in pyproject.toml and use the generated file here:
        # from ._version import version as __version__
    except PackageNotFoundError:
        warnings.warn('Could not find pyleco version, it does not seem to be installed. '
                      'Either install it (editable or full) or install setuptools_scm')
        __version__ = '0.0.0'
    finally:
        del version, PackageNotFoundError



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
import threading
from time import sleep

import pytest

from pyleco.coordinators.coordinator import Coordinator
from pyleco.actors.actor import Actor
from pyleco.directors.director import Director


# Constants
PORT = 60004


def start_coordinator(namespace: str, port: int, coordinators=None, **kwargs):
    with Coordinator(namespace=namespace, port=port, **kwargs) as coordinator:
        coordinator.routing(coordinators=coordinators)


class FakeInstrument:
    _prop1 = 5

    def __init__(self):
        pass

    def connect(self):
        pass

    @property
    def constant(self):
        return 7

    @property
    def prop1(self):
        return self._prop1

    @prop1.setter
    def prop1(self, value):
        self._prop1 = value

    def triple(self, factor: float = 1, factor2: float = 1) -> float:
        return factor * factor2 * 3


def start_actor(event: threading.Event):
    actor = Actor("actor", FakeInstrument, port=PORT)

    def binary_method_manually() -> None:
        """Receive binary data and return it. Do all the binary things manually."""
        payload = actor.current_message.payload[1:]
        try:
            actor.additional_response_payload = [payload[0] * 2]
        except IndexError:
            pass

    def binary_method_created(additional_payload: list[bytes]) -> tuple[None, list[bytes]]:
        """Receive binary data and return it. Create binary method by registering it."""
        return None, [additional_payload[0] * 2]

    actor.register_rpc_method(binary_method_manually)
    actor.register_binary_rpc_method(
        binary_method_created, accept_binary_input=True, return_binary_output=True
    )
    actor.connect()
    actor.rpc.method()(actor.device.triple)
    actor.register_device_method(actor.device.triple)
    actor.listen(event)
    actor.disconnect()


@pytest.fixture(scope="module")
def director():
    """A leco setup."""
    glog = logging.getLogger()
    glog.setLevel(logging.DEBUG)
    # glog.addHandler(logging.StreamHandler())
    log = logging.getLogger("test")
    stop_event = threading.Event()
    threads = []
    threads.append(threading.Thread(target=start_coordinator,
                                    kwargs=dict(namespace="N1", port=PORT)))
    threads.append(threading.Thread(target=start_actor, kwargs=dict(event=stop_event)))
    for thread in threads:
        thread.daemon = True
        thread.start()
    sleep(1)
    director = Director(actor="actor", port=PORT)
    yield director
    log.info("Tearing down")
    stop_event.set()
    director.shut_down_actor(actor="COORDINATOR")
    for thread in threads:
        thread.join(0.5)


def test_get_property(director: Director):
    assert director.get_parameters("constant") == {"constant": 7}


def test_change_property(director: Director):
    start = director.get_parameters(["prop1"])["prop1"]
    director.set_parameters({"prop1": start + 3})
    assert director.get_parameters(["prop1"])["prop1"] == start + 3


def test_call_action_arg(director: Director):
    assert director.call_action("triple", 5) == 15


def test_call_action_kwarg(director: Director):
    assert director.call_action(action="triple", factor=5) == 15


def test_call_action_arg_and_kwarg(director: Director):
    assert director.call_action("triple", 2, factor2=5) == 30


def test_method_via_rpc(director: Director):
    assert director.ask_rpc(method="triple", factor=5) == 15


def test_method_via_rpc2(director: Director):
    assert director.ask_rpc(method="triple", factor=2, factor2=5) == 30


def test_device_method_via_rpc(director: Director):
    assert director.ask_rpc(method="device.triple", factor=5) == 15


def test_binary_data_transfer(director: Director):
    assert director.ask_rpc(
        method="binary_method_manually",
        additional_payload=[b"123"],
        extract_additional_payload=True,
    ) == (None, [b"123123"])


def test_binary_data_transfer_created(director: Director):
    assert director.ask_rpc(
        method="binary_method_created", additional_payload=[b"123"], extract_additional_payload=True
    ) == (None, [b"123123"])



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
These classes show which methods have to be available via RPC in order to comply with LECO message
definitions.

A combination of type checking and unit tests can test the compliance with LECO definitions.

For example, if you want to verify, that the class `Actor` fulfills the requirements for a
Component, for an Actor which supports Polling and setting the log level, you may use the following
tests.

For a static test, that all the methods are present with the correct types, the following works:

.. code::

    class ExtendedActorProtocol(ExtendedComponentProtocol, PollingActorProtocol, Protocol):
        "Combine all required Protocols for the class under test."
        pass

    def static_test_methods_are_present():
        def testing(component: ExtendedActorProtocol):
            pass
        testing(Actor(name="test", cls=FantasyInstrument))

For unit test, that all the necessary methods are reachable via RPC, the following works:

.. code::

    protocol_methods = [m for m in dir(ExtendedActorProtocol) if not m.startswith("_")]

    @pytest.fixture
    def component_methods(actor: Actor):
        response = actor.rpc.process_request(
            '{"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}')
        result = actor.rpc_generator.get_result_from_response(response)  # type: ignore
        return result.get('methods')

    @pytest.mark.parametrize("method", protocol_methods)
    def test_method_is_available(component_methods, method):
        for m in component_methods:
            if m.get('name') == method:
                return
        raise AssertionError(f"Method {method} is not available.")
"""

from __future__ import annotations
try:
    from enum import StrEnum  # type: ignore
except ImportError:
    # For python<3.11
    from enum import Enum

    class StrEnum(str, Enum):  # type: ignore
        pass

from typing import Any, Iterable, Optional, Protocol, Sequence


class ComponentProtocol(Protocol):
    """Any Component of the LECO protocol."""

    def pong(self) -> None:
        """Respond to any request."""
        return  # always succeeds.


class LogLevels(StrEnum):
    """Log levels for :meth:`ExtendedComponentProtocol.set_log_level` method."""
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class ExtendedComponentProtocol(ComponentProtocol, Protocol):
    """A Component which supports more features."""

    def set_log_level(self, level: LogLevels) -> None: ...

    def shut_down(self) -> None: ...


class CoordinatorProtocol(ComponentProtocol, Protocol):
    """A command protocol Coordinator."""

    def sign_in(self) -> None: ...

    def sign_out(self) -> None: ...

    def coordinator_sign_in(self) -> None: ...

    def coordinator_sign_out(self) -> None: ...

    def add_nodes(self, nodes: dict[str, str]) -> None: ...

    def send_nodes(self) -> dict[str, str]: ...

    def record_components(self, components: list[str]) -> None: ...

    def send_local_components(self) -> list[str]: ...

    def send_global_components(self) -> dict[str, list[str]]: ...

    def remove_expired_addresses(self, expiration_time: float) -> None: ...


class ActorProtocol(ComponentProtocol, Protocol):
    """An Actor Component."""

    def get_parameters(self, parameters: Iterable[str]) -> dict[str, Any]: ...

    def set_parameters(self, parameters: dict[str, Any]) -> None: ...

    def call_action(self, action: str, args: Optional[Sequence[Any]] = None,
                    kwargs: Optional[dict[str, Any]] = None) -> Any: ...


class PollingActorProtocol(ActorProtocol, Protocol):
    """An Actor which allows regular polling."""

    def start_polling(self, polling_interval: Optional[float]) -> None: ...

    def set_polling_interval(self, polling_interval: float) -> None: ...

    def get_polling_interval(self) -> float: ...

    def stop_polling(self) -> None: ...


class LockingActorProtocol(ActorProtocol, Protocol):
    """An Actor which allows to lock the device or channels of the device."""

    def lock(self, resource: Optional[str] = None) -> bool: ...

    def unlock(self, resource: Optional[str] = None) -> None: ...

    def force_unlock(self, resource: Optional[str] = None) -> None: ...



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from unittest.mock import call, MagicMock

import pytest

from pyleco.test import FakeContext, assert_response_is_result, handle_request_message
from pyleco.management.starter import sanitize_tasks, Starter, Status
from pyleco.utils.events import SimpleEvent


@pytest.fixture
def starter() -> Starter:
    starter = Starter(context=FakeContext())
    return starter


class FakeThread:
    def __init__(self, target=None, alive=False, *args, **kwargs) -> None:
        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._alive = alive

    def start(self) -> None:
        self._alive = True

    def is_alive(self) -> bool:
        return self._alive

    def join(self, timeout=None) -> None:
        return


@pytest.mark.parametrize("tasks", (None, [], (), "string", ["abc", "def"]))
def test_sanitize_tasks(tasks):
    sanitized = sanitize_tasks(tasks)
    assert isinstance(sanitized, (tuple, list))
    for t in sanitized:
        assert isinstance(t, str)


@pytest.mark.parametrize("tasks, invalid_task_name", (
        (5, 5),
        (("valid", 6), 6),
        ([["list"], "abc"], "['list']")),
)
def test_invalid_tasks(tasks, invalid_task_name, caplog):
    assert sanitize_tasks(tasks) == ()
    assert caplog.messages == [f"Invalid task name '{invalid_task_name}' received."]



def test_init(starter: Starter):
    assert starter.started_tasks == {}
    assert starter.threads == {}


@pytest.mark.parametrize("pre, post", (
        (Status.STARTED | Status.INSTALLED, Status.STARTED | Status.INSTALLED),
        (None, Status.INSTALLED),  # not yet in the dict
        (Status.STARTED, Status.STARTED | Status.INSTALLED),
        (Status.STARTED | Status.RUNNING, Status.STARTED | Status.RUNNING | Status.INSTALLED),
        (Status.STOPPED, Status.INSTALLED),
))
def test_install_task(starter: Starter, pre: Status, post: Status):
    if pre is not None:
        starter.started_tasks["test"] = pre
    starter.install_task("test")
    assert starter.started_tasks["test"] == post


def test_install_tasks(starter: Starter):
    # arrange
    starter.install_task = MagicMock()  # type: ignore[method-assign]
    # act
    handle_request_message(starter, "install_tasks", names=["a", "b"])
    # assert
    assert starter.install_task.call_args_list == [call("a"), call("b")]
    assert_response_is_result(starter)


@pytest.mark.parametrize("pre, post", (
        (Status.RUNNING | Status.INSTALLED, Status.RUNNING),
        (None, Status.STOPPED),  # not yet in the dict
        (Status.RUNNING, Status.RUNNING),
        (Status.STOPPED, Status.STOPPED),
))
def test_uninstall_task(starter: Starter, pre: Status, post: Status):
    if pre is not None:
        starter.started_tasks["test"] = pre
    starter.uninstall_task("test")
    assert starter.started_tasks.get("test") == post


def test_uninstall_tasks(starter: Starter):
    # arrange
    starter.uninstall_task = MagicMock()  # type: ignore[method-assign]
    # act
    handle_request_message(starter, "uninstall_tasks", names=["a", "b"])
    # assert
    assert starter.uninstall_task.call_args_list == [call("a"), call("b")]
    assert_response_is_result(starter)


class Test_status_tasks:
    @pytest.fixture
    def status(self, starter: Starter) -> dict[str, Status]:
        starter.threads["SR"] = FakeThread(alive=True)  # type: ignore
        starter.threads["S"] = FakeThread()  # type: ignore
        starter.threads["NS"] = FakeThread(alive=True)  # type: ignore
        starter.started_tasks = {
            "SR": Status.STARTED | Status.RUNNING,
            "S": Status.STARTED | Status.RUNNING,
            "NS": Status.STARTED,
        }
        self.starter = starter
        handle_request_message(starter, "status_tasks", names=["unknown"])
        result = assert_response_is_result(starter)
        return result

    def test_started_running(self, status):
        """Test that a running task remains running."""
        assert status["SR"] == Status.STARTED | Status.RUNNING

    def test_started_not_running(self, status):
        """Test that a stopped thread is not running anymore."""
        assert status["S"] == Status.STARTED

    def test_newly_started_is_also_running(self, status):
        """Test that a newly started (last time not running) is now running."""
        assert status["NS"] == Status.STARTED | Status.RUNNING

    def test_unknown_is_marked_stopped(self, status):
        assert status["unknown"] == Status.STOPPED

    def test_stopped_running_thread_is_removed(self, status):
        assert "S" not in self.starter.threads.keys()

    def test_stopped_causes_log_entry(self, status, caplog: pytest.LogCaptureFixture):
        assert "Thread 'S' stopped unexpectedly." in [
            record.message for record in caplog.get_records(when="setup")
        ]


class Test_check_installed_tasks:
    @pytest.fixture
    def starter_cit(self, starter: Starter) -> Starter:
        starter.start_task = MagicMock()  # type: ignore[method-assign]
        starter.started_tasks = {
            "IR": Status.INSTALLED | Status.RUNNING,
            "INR": Status.INSTALLED,
            "SR": Status.STARTED | Status.RUNNING,
            "SNR": Status.STARTED,  # not running, should not be started as it is not installed.
        }
        starter.check_installed_tasks()
        return starter

    def test_start_installed_but_not_running_task(self, starter_cit: Starter):
        """Test, that only the installed (and not running) task is started."""
        starter_cit.start_task.assert_called_once_with("INR")  # type: ignore[attr-defined]


class Test_start_task:
    def test_already_started_task(self, starter: Starter):
        # arrange
        starter.started_tasks["t1"] = Status.STARTED
        starter.threads["t1"] = FakeThread(alive=True)  # type: ignore
        starter.events["t1"] = SimpleEvent()  # type: ignore
        # act
        starter.start_task("t1")
        assert Status.RUNNING in Status(starter.started_tasks["t1"])


def test_start_tasks(starter: Starter):
    # arrange
    starter.start_task = MagicMock()  # type: ignore[method-assign]
    # act
    handle_request_message(starter, "start_tasks", names=["a", "b"])
    # assert
    assert starter.start_task.call_args_list == [call("a"), call("b")]
    assert_response_is_result(starter)


class Test_stop_task:
    def test_stop_not_existing_task(self, starter: Starter):
        starter.stop_task("whatever")

    def test_stop_existing_running_task(self, starter: Starter):
        # arrange
        starter.started_tasks["t1"] = Status.STARTED
        starter.threads["t1"] = FakeThread(alive=True)  # type: ignore
        event = starter.events["t1"] = SimpleEvent()  # type: ignore
        # act
        starter.stop_task("t1")
        assert "t1" not in starter.threads
        assert "t1" not in starter.started_tasks
        assert event.is_set() is True

    def test_stop_removed_task(self, starter: Starter):
        # arrange
        try:
            del starter.threads["t1"]
        except KeyError:
            pass
        starter.started_tasks["t1"] = Status.STARTED
        # act
        starter.stop_task("t1")
        assert "t1" not in starter.threads
        assert "t1" not in starter.started_tasks


def test_stop_tasks(starter: Starter):
    # arrange
    starter.stop_task = MagicMock()  # type: ignore[method-assign]
    # act
    handle_request_message(starter, "stop_tasks", names=["a", "b"])
    # assert
    assert starter.stop_task.call_args_list == [call("a"), call("b")]
    assert_response_is_result(starter)


def test_restart_tasks(starter: Starter):
    starter.start_task = MagicMock()  # type: ignore[method-assign]
    starter.stop_task = MagicMock()  # type: ignore[method-assign]
    starter.restart_tasks(["a", "b"])
    assert starter.stop_task.call_args_list == [call("a"), call("b")]
    assert starter.start_task.call_args_list == [call("a"), call("b")]


def test_stop_all_tasks(starter: Starter):
    # arrange
    starter.started_tasks["t1"] = Status.STARTED
    starter.threads["t1"] = FakeThread(alive=True)  # type: ignore
    event = starter.events["t1"] = SimpleEvent()  # type: ignore
    # act
    starter.stop_all_tasks()
    assert "t1" not in starter.threads
    assert "t1" not in starter.started_tasks
    assert event.is_set() is True


def test_list_tasks_failing(starter: Starter):
    starter.directory = "/abcdefghijklmno"
    assert starter.list_tasks() == []



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.core.message import Message, MessageTypes
from pyleco.core.internal_protocols import CommunicatorProtocol
from pyleco.test import FakeCommunicator
from pyleco.json_utils.errors import JSONRPCError

cid = b"conversation_id;"

# Test the utility methods of the CommunicatorProtocol


@pytest.fixture
def communicator() -> CommunicatorProtocol:
    return FakeCommunicator(name="communicator")


def test_full_name_without_namespace(communicator: FakeCommunicator):
    communicator.namespace = None
    assert communicator.full_name == "communicator"


def test_full_name_with_namespace(communicator: FakeCommunicator):
    communicator.namespace = "N1"
    assert communicator.full_name == "N1.communicator"


def test_send(communicator: FakeCommunicator):
    kwargs = dict(receiver="rec", message_type=MessageTypes.JSON, data=[4, 5], conversation_id=cid)
    communicator.send(**kwargs)  # type: ignore
    assert communicator._s[0] == Message(sender="communicator", **kwargs)  # type: ignore


class Test_ask:
    response = Message(receiver="communicator", sender="rec", conversation_id=cid)

    @pytest.fixture
    def communicator_asked(self, communicator: FakeCommunicator):
        communicator._r = [self.response]
        return communicator

    def test_sent(self, communicator_asked: FakeCommunicator):
        communicator_asked.ask(receiver="rec", conversation_id=cid)
        assert communicator_asked._s == [Message(receiver="rec", sender="communicator",
                                                 conversation_id=cid)]

    def test_read(self, communicator_asked: FakeCommunicator):
        response = communicator_asked.ask(receiver="rec", conversation_id=cid)
        assert response == self.response


class Test_interpret_rpc_response:
    def test_valid_message(self, communicator: FakeCommunicator):
        message = Message(receiver="rec", data={"jsonrpc": "2.0", "result": 6.0, "id": 7})
        assert communicator.interpret_rpc_response(message) == 6.0

    def test_error(self, communicator: FakeCommunicator):
        message = Message(receiver="rec", data={"jsonrpc": "2.0",
                                                "error": {"code": -1, "message": "abc"}, "id": 7})
        with pytest.raises(JSONRPCError):
            communicator.interpret_rpc_response(message)

    def test_json_binary_response(self, communicator: FakeCommunicator):
        message = Message(
            receiver="rec",
            data={"jsonrpc": "2.0", "result": None, "id": 7},
            additional_payload=[b"abcd", b"efgh"],
        )
        assert communicator.interpret_rpc_response(message, extract_additional_payload=True) == (
            None,
            [
                b"abcd",
                b"efgh",
            ],
        )

    def test_ignore_additional_payload_if_not_desired(self, communicator: FakeCommunicator):
        message = Message(
            receiver="rec",
            data={"jsonrpc": "2.0", "result": None, "id": 7},
            additional_payload=[b"abcd"],
        )
        assert (
            communicator.interpret_rpc_response(message, extract_additional_payload=False) is None
        )

    def test_without_additional_payload_return_empty_list(self, communicator: FakeCommunicator):
        message = Message(
            receiver="rec",
            data={"jsonrpc": "2.0", "result": None, "id": 7},
        )
        assert communicator.interpret_rpc_response(message, extract_additional_payload=True) == (
            None,
            [],
        )

    def test_json_value_and_binary_payload(self, communicator: FakeCommunicator):
        message = Message(
            receiver="rec",
            data={"jsonrpc": "2.0", "result": 6, "id": 7},
            additional_payload=[b"abcd"],
        )
        assert communicator.interpret_rpc_response(message, extract_additional_payload=True) == (
            6,
            [b"abcd"],
        )


class Test_ask_rpc:
    response = Message(receiver="communicator", sender="rec", conversation_id=cid,
                       message_type=MessageTypes.JSON,
                       data={
                           "jsonrpc": "2.0",
                           "result": 5,
                           "id": 1,
                           })

    @pytest.fixture
    def communicator_asked(self, communicator: FakeCommunicator):
        communicator._r = [self.response]
        return communicator

    def test_sent(self, communicator_asked: FakeCommunicator):
        communicator_asked.ask_rpc(receiver="rec", method="test_method", par1=5)
        sent = communicator_asked._s[0]
        assert communicator_asked._s == [Message(receiver="rec", sender="communicator",
                                                 conversation_id=sent.conversation_id,
                                                 message_type=MessageTypes.JSON,
                                                 data={
                                                     "jsonrpc": "2.0",
                                                     "method": "test_method",
                                                     "id": 1,
                                                     "params": {'par1': 5},
                                                 })]

    def test_sent_with_additional_payload(self, communicator_asked: FakeCommunicator):
        communicator_asked.ask_rpc(
            receiver="rec", method="test_method", par1=5, additional_payload=[b"12345"]
        )
        sent = communicator_asked._s[0]
        assert communicator_asked._s == [
            Message(
                receiver="rec",
                sender="communicator",
                conversation_id=sent.conversation_id,
                message_type=MessageTypes.JSON,
                data={
                    "jsonrpc": "2.0",
                    "method": "test_method",
                    "id": 1,
                    "params": {"par1": 5},
                },
                additional_payload=[b"12345"],
            )
        ]

    def test_read(self, communicator_asked: FakeCommunicator):
        result = communicator_asked.ask_rpc(receiver="rec", method="test_method", par1=5)
        assert result == 5



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from unittest.mock import MagicMock

import pytest

from pyleco.core import VERSION_B
from pyleco.core.message import Message, MessageTypes
from pyleco.core.serialization import serialize_data
from pyleco.json_utils.errors import JSONRPCError, NOT_SIGNED_IN, NODE_UNKNOWN

from pyleco.utils.communicator import Communicator
from pyleco.test import FakeSocket, FakeContext


cid = b"conversation_id;"
header = b"".join((cid, b"\x00" * 3, b"\x01"))


message_tests = (
    ({'receiver': "broker", 'data': [["GET", [1, 2]], ["GET", 3]], 'sender': 's',
      'message_type': MessageTypes.JSON},
     [VERSION_B, b"broker", b"s", header, serialize_data([["GET", [1, 2]], ["GET", 3]])]),
    ({'receiver': "someone", 'conversation_id': cid, 'sender': "ego", 'message_id': b"mid"},
     [VERSION_B, b'someone', b'ego', b'conversation_id;mid\x00']),
    ({'receiver': "router", 'sender': "origin"},
     [VERSION_B, b"router", b"origin", header[:-1] + b"\x00"]),
)


def fake_time():
    return 0


def fake_randbytes(n):
    return b"\01" * n


def fake_generate_cid():
    return cid


@pytest.fixture()
def fake_cid_generation(monkeypatch):
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id", fake_generate_cid)


# intercom
class FakeCommunicator(Communicator):
    def open(self, context=None):
        super().open(context=FakeContext())  # type: ignore


@pytest.fixture()
def communicator() -> Communicator:
    communicator = FakeCommunicator(name="Test")
    communicator._last_beat = float("inf")
    return communicator


def test_name():
    c = FakeCommunicator(name="Test")
    assert c.name == "Test"


def test_auto_open():
    c = FakeCommunicator(name="Test", auto_open=True)
    assert isinstance(c.socket, FakeSocket)


def test_context_manager_opens_connection():
    class FK2(FakeCommunicator):
        def __init__(self, **kwargs):
            super().__init__(auto_open=False, **kwargs)

        def sign_in(self):
            pass
    with FK2(name="Test") as c:
        assert isinstance(c.socket, FakeSocket)


class Test_close:
    @pytest.fixture
    def closed_communicator(self, communicator: Communicator, fake_cid_generation):
        message = Message("Test", "COORDINATOR", message_type=MessageTypes.JSON,
                          conversation_id=cid, data={
                              "jsonrcp": "2.0", "result": None, "id": 1,
                              })
        communicator.socket._r = [message.to_frames()]  # type: ignore
        communicator.close()
        return communicator

    def test_socket_closed(self, closed_communicator: Communicator):
        assert closed_communicator.socket.closed is True

    def test_signed_out(self, closed_communicator: Communicator):
        sign_out_message = Message.from_frames(*closed_communicator.socket._s.pop())  # type: ignore  # noqa
        assert sign_out_message == Message(
            "COORDINATOR",
            "Test",
            conversation_id=sign_out_message.conversation_id,
            message_type=MessageTypes.JSON,
            data={'jsonrpc': "2.0", 'method': "sign_out", "id": 1}
        )

    def test_no_error_without_socket(self):
        communicator = FakeCommunicator("Test", auto_open=False)
        communicator.close()
        # no error raised


def test_reset(communicator: Communicator):
    communicator.close = MagicMock()  # type: ignore
    communicator.open = MagicMock()  # type: ignore
    # act
    communicator.reset()
    # assert
    communicator.close.assert_called_once()
    communicator.open.assert_called_once()


@pytest.mark.parametrize("kwargs, message", message_tests)
def test_communicator_send(communicator: Communicator, kwargs, message, monkeypatch,
                           fake_cid_generation):
    monkeypatch.setattr("pyleco.utils.communicator.perf_counter", fake_time)
    communicator.send(**kwargs)
    assert communicator.socket._s.pop() == message  # type: ignore


def test_poll(communicator: Communicator):
    assert communicator.poll() == 0


class Test_ask_message:
    request = Message(receiver=b"N1.receiver", data="whatever", conversation_id=cid)
    response = Message(receiver=b"N1.Test", sender=b"N1.receiver", data=["xyz"],
                       message_type=MessageTypes.JSON,
                       conversation_id=cid)

    def test_ignore_ping(self, communicator: Communicator):
        ping_message = Message(receiver=b"N1.Test", sender=b"N1.COORDINATOR",
                               message_type=MessageTypes.JSON,
                               data={"id": 0, "method": "pong", "jsonrpc": "2.0"})
        communicator.socket._r = [  # type: ignore
            ping_message.to_frames(),
            self.response.to_frames()]
        communicator.ask_message(self.request)
        assert communicator.socket._s == [self.request.to_frames()]

    def test_sign_in(self, communicator: Communicator, fake_cid_generation):
        communicator.sign_in = MagicMock()  # type: ignore
        not_signed_in = Message(receiver="N1.Test", sender="N1.COORDINATOR",
                                message_type=MessageTypes.JSON,
                                conversation_id=cid,
                                data={"id": None,
                                      "error": NOT_SIGNED_IN.model_dump(),
                                      "jsonrpc": "2.0"},
                                )
        communicator.socket._r = [  # type: ignore
            not_signed_in.to_frames(),
            self.response.to_frames()]
        response = communicator.ask_message(self.request)
        # assert that the message is sent once
        assert communicator.socket._s.pop(0) == self.request.to_frames()  # type: ignore
        # assert that it tries to sign in
        communicator.sign_in.assert_called()
        # assert that the message is called a second time
        assert communicator.socket._s == [self.request.to_frames()]
        # assert that the correct response is returned
        assert response == self.response

    def test_ignore_wrong_response(self, communicator: Communicator,
                                   caplog: pytest.LogCaptureFixture):
        """A wrong response should not be returned."""
        caplog.set_level(10)
        m = Message(receiver="whatever", sender="s", message_type=MessageTypes.JSON,
                    data={'jsonrpc': "2.0"}).to_frames()
        communicator.socket._r = [m, self.response.to_frames()]  # type: ignore
        assert communicator.ask_message(self.request) == self.response

    def test_sign_in_fails_several_times(self, communicator: Communicator, fake_cid_generation):
        not_signed_in = Message(receiver="communicator", sender="N1.COORDINATOR",
                                message_type=MessageTypes.JSON,
                                data={"id": None,
                                      "error": NOT_SIGNED_IN.model_dump(),
                                      "jsonrpc": "2.0"},
                                ).to_frames()
        communicator.sign_in = MagicMock()  # type: ignore
        communicator.socket._r = [not_signed_in, not_signed_in]  # type: ignore
        with pytest.raises(ConnectionRefusedError):
            communicator.ask_message(self.request)

    @pytest.mark.xfail(True, reason="Unsure whether it should work that way.")
    def test_ask_message_with_error(self, communicator: Communicator):
        response = Message(receiver="communicator", sender="N1.COORDINATOR",
                        message_type=MessageTypes.JSON, conversation_id=cid,
                        data={"id": None,
                                "error": NODE_UNKNOWN.model_dump(),
                                "jsonrpc": "2.0"},
                        )
        communicator.socket._r = [response.to_frames()]  # type: ignore
        with pytest.raises(JSONRPCError, match=NODE_UNKNOWN.message):
            communicator.ask_message(Message("receiver", conversation_id=cid))


def test_ask_rpc(communicator: Communicator, fake_cid_generation):
    received = Message(receiver=b"N1.Test", sender=b"N1.receiver",
                       conversation_id=cid)
    received.payload = [b"""{"jsonrpc": "2.0", "result": 123.45, "id": "1"}"""]
    communicator.socket._r = [received.to_frames()]  # type: ignore
    response = communicator.ask_rpc(receiver="N1.receiver", method="test_method", some_arg=4)
    assert communicator.socket._s == [
        Message(b'N1.receiver', b'Test',
                conversation_id=cid,
                message_type=MessageTypes.JSON,
                data={"id": 1, "method": "test_method",
                      "params": {"some_arg": 4}, "jsonrpc": "2.0"}).to_frames()]
    assert response == 123.45


def test_communicator_sign_in(fake_cid_generation, communicator: Communicator):
    communicator.socket._r = [  # type: ignore
        Message(b"N2.n", b"N2.COORDINATOR",
                conversation_id=cid, message_type=MessageTypes.JSON,
                data={"id": 1, "result": None, "jsonrpc": "2.0"}).to_frames()]
    communicator.sign_in()
    assert communicator.namespace == "N2"


def test_get_capabilities(communicator: Communicator, fake_cid_generation):
    communicator.socket._r = [  # type: ignore
        Message("communicator", "sender", conversation_id=cid,
                message_type=MessageTypes.JSON,
                data={"id": 1, "result": 6, "jsonrpc": "2.0"}
                ).to_frames()
    ]
    result = communicator.get_capabilities(receiver="rec")
    sent = Message.from_frames(*communicator.socket._s.pop())  # type: ignore
    assert sent.data == {"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}
    assert result == 6


def test_ask_json(communicator: Communicator, fake_cid_generation):
    response = Message("communicator", sender="rec", conversation_id=cid,
                       data="super response")
    communicator.socket._r = [response.to_frames()]  # type: ignore
    json_string = "[5, 6.7]"
    # act
    result = communicator.ask_json(receiver="rec", json_string=json_string)
    # assert
    sent = Message.from_frames(*communicator.socket._s.pop())  # type: ignore
    assert sent.data == [5, 6.7]
    assert result == b"super response"



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from enum import IntFlag
from importlib import import_module, reload
import logging
import os
from os import path
import sys
import threading
from typing import Any, Optional, Union

if __name__ != "__main__":
    from ..utils.message_handler import MessageHandler
    from ..utils.parser import parser, parse_command_line_parameters
else:  # pragma: no cover
    from pyleco.utils.message_handler import MessageHandler
    from pyleco.utils.parser import parser, parse_command_line_parameters


log = logging.getLogger("starter")
StrFormatter = logging.Formatter("%(asctime)s\t%(levelname)s\t%(name)s\t%(message)s")


modules: dict[str, Any] = {}  # A dictionary of the task modules


def sanitize_tasks(
    tasks: Optional[Union[list[str], tuple[str, ...], str]],
) -> Union[tuple[str, ...], list[str]]:
    """Ensure that the tasks are a list of tasks."""
    if tasks is None:
        return ()
    if not isinstance(tasks, (list, tuple)):
        tasks = (tasks,)
    for task in tasks:
        if not isinstance(task, str):
            log.error(f"Invalid task name '{task}' received.")
            return ()
    return tasks


class Status(IntFlag):
    STOPPED = 0
    RUNNING = 1  # currently running
    STARTED = 2  # has been started and should be running
    INSTALLED = 4  # check regularly, whether it is running, restart if not running anymore


class Starter(MessageHandler):
    """Listen to communication and start tasks as required.

    The Starter can start functions called `task` with the following signature
    ``task(stop_event: threading.Event) -> None: ...``.
    Whenever a task should be started, the Starter looks in :attr:`directory` for a module with that
    name and loads it.
    If the module has been loaded already, it is reloaded to get the newest version.
    Attention: dependencies are not reloaded!
    Then it starts the method `task` of that module in a separate thread.

    When the Starter stops a task, it sets the corresponding `threading.Event`.

    The first str line of a module is the modules description, available from :meth:`list_tasks`.

    If you write your own task modules, make sure, that you react in a reasonable time to a stop
    event.
    You can use `while stop_event.wait(timeout)` to execute something regularly.
    Pyleco Actors should be able to consume a `threading.Event` in order to be compatible.

    .. code::
        starter = Starter("starter")
        starter.listen()

    :param str directory: Absolute path to the directory with the tasks modules.
    :param tasks: List of task names to execute on startup.
    """

    def __init__(
        self,
        name: str = "starter",
        directory: Optional[str] = None,
        tasks: Optional[list[str]] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.threads: dict[str, threading.Thread] = {}  # List of threads
        self.events: dict[str, threading.Event] = {}  # Events to stop the threads.
        self.started_tasks: dict[str, int | Status] = {}  # A list of all tasks started
        if directory is not None:
            self.directory = path.normpath(directory)
            head, tail = path.split(self.directory)
            sys.path.append(head)
            self.folder_name = tail
        else:
            # TODO remove?
            self.directory = "test_tasks"
            self.folder_name = "test_tasks"

        log.info(f"Starter started with tasks in folder '{self.directory}'.")
        self.start_tasks(tasks or ())

    def register_rpc_methods(self) -> None:
        super().register_rpc_methods()
        self.register_rpc_method(self.start_tasks)
        self.register_rpc_method(self.stop_tasks)
        self.register_rpc_method(self.restart_tasks)
        self.register_rpc_method(self.install_tasks)
        self.register_rpc_method(self.list_tasks)
        self.register_rpc_method(self.status_tasks)
        self.register_rpc_method(self.uninstall_tasks)

    def _listen_close(self, waiting_time: Optional[int] = None) -> None:
        """Close the listening loop."""
        super()._listen_close(waiting_time=waiting_time)
        self.stop_all_tasks()
        log.info("Starter stopped.")

    def stop_all_tasks(self) -> None:
        self.started_tasks = {}
        keys = list(self.threads.keys())
        for name in keys:
            # set all stop signals
            self.events[name].set()
        for name in keys:
            self.wait_for_stopped_thread(name)

    def heartbeat(self) -> None:
        """Check installed tasks at heartbeating."""
        super().heartbeat()
        self.check_installed_tasks()

    def start_tasks(self, names: Union[list[str], tuple[str, ...]]) -> None:
        for name in sanitize_tasks(names):
            self.start_task(name)

    def start_task(self, name: str) -> None:
        """Start the `Task` object in a script with `name` in a separate thread."""
        if name in self.threads.keys() and self.threads[name].is_alive():
            log.error(f"Task '{name}' is already running.")
            self.started_tasks[name] |= Status.RUNNING
        else:
            log.info(f"Starting task '{name}'.")
            self.started_tasks[name] = self.started_tasks.get(name, 0) | Status.STARTED
            try:
                if name in modules.keys():
                    modules[name] = script = reload(modules[name])
                else:
                    modules[name] = script = import_module(f"{self.folder_name}.{name}")
            except Exception as exc:
                log.exception(f"Loading task '{name}' failed.", exc_info=exc)
                return
            self.events[name] = threading.Event()
            try:
                self.threads[name] = thread = threading.Thread(
                    target=script.task, args=(self.events[name],), daemon=True
                )
            except Exception as exc:
                log.exception(f"Creation of task '{name}' failed.", exc_info=exc)
                return
            thread.start()

    def stop_tasks(self, names: Union[list[str], tuple[str, ...]]) -> None:
        for name in sanitize_tasks(names):
            self.stop_task(name)

    def stop_task(self, name: str) -> None:
        """Stop a task and don't restart it, if it was installed."""
        try:
            del self.started_tasks[name]
        except KeyError:
            pass  # Not present
        if name not in self.threads.keys():
            return
        log.info(f"Stopping task '{name}'.")
        self.events[name].set()
        self.wait_for_stopped_thread(name)

    def wait_for_stopped_thread(self, name: str) -> None:
        thread = self.threads[name]
        thread.join(timeout=2)
        if thread.is_alive():
            log.warning(f"Task '{name}' did not stop in time!")
            # TODO add possibility to stop thread otherwise.
        try:
            del self.threads[name]
        except Exception as exc:
            log.exception(f"Deleting task '{name}' failed", exc_info=exc)

    def restart_tasks(self, names: Union[list[str], tuple[str, ...]]) -> None:
        for name in sanitize_tasks(names):
            self.stop_task(name)
            self.start_task(name)

    def install_tasks(self, names: Union[list[str], tuple[str, ...]]) -> None:
        for name in sanitize_tasks(names):
            self.install_task(name)

    def install_task(self, name: str) -> None:
        """Add tasks to the installed list."""
        log.info(f"Install task '{name}'.")
        self.started_tasks[name] = self.started_tasks.get(name, 0) | Status.INSTALLED

    def uninstall_tasks(self, names: Union[list[str], tuple[str, ...]]) -> None:
        for name in sanitize_tasks(names):
            self.uninstall_task(name)

    def uninstall_task(self, name: str) -> None:
        """Uninstalls a task without stopping it, if it is already running."""
        self.started_tasks[name] = self.started_tasks.get(name, 0) & ~Status.INSTALLED

    def status_tasks(self, names: Optional[list[str]] = None) -> dict[str, Status]:
        """Enumerate the status of the started/running tasks and keep the records clean.

        :param list names: List of tasks to look for.
        """
        ret_data = {} if names is None else {key: Status.STOPPED for key in names}
        for key in list(self.threads.keys()):
            if self.threads[key].is_alive():
                self.started_tasks[key] |= Status.RUNNING
            else:
                self.started_tasks[key] = self.started_tasks.get(key, 0) & ~Status.RUNNING
                del self.threads[key]
                log.warning(f"Thread '{key}' stopped unexpectedly.")
        ret_data.update(self.started_tasks)  # type: ignore
        return ret_data

    def list_tasks(self) -> list[dict[str, str]]:
        """List all tasks (with name and tooltip) available in the folder."""
        try:
            filenames = os.listdir(self.directory)
        except FileNotFoundError:
            log.error(f"Task folder '{self.directory}' not found.")
            return []
        tasks = []
        for name in filenames:
            if name.endswith(".py") and not name == "__init__.py":
                with open(f"{self.directory}/{name}", "r") as file:
                    # Search for the first line with triple quotes
                    for i in range(10):
                        if file.readline().strip() == '"""':
                            break
                    tooltip = file.readline()  # first line after line with triple quotes
                tasks.append({"name": name.replace(".py", ""), "tooltip": tooltip})
        log.debug(f"Tasks found: {tasks}.")
        return tasks

    def check_installed_tasks(self) -> None:
        """Check whether installed tasks are running."""
        self.status_tasks()
        for task, s in self.started_tasks.items():
            if s & Status.INSTALLED and not s & Status.RUNNING:
                log.info(f"Starting installed task '{task}' with status {s}.")
                self.start_task(task)


def main() -> None:
    parser.add_argument("tasks", nargs="*", help="Tasks to execute at startup.")
    parser.add_argument(
        "-d",
        "--directory",
        help="set the directory to search for tasks, do not add a trailing slash",
    )

    gLog = logging.getLogger()  # print all log entries!
    if not gLog.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(StrFormatter)
        gLog.addHandler(handler)
    kwargs = parse_command_line_parameters(
        parser=parser, parser_description="Start tasks as required.", logger=gLog
    )

    starter = Starter(log=gLog, **kwargs)
    starter.listen()


if __name__ == "__main__":  # pragma: no cover
    main()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from typing import Any, Iterable, Optional, Sequence, Union

from .core.message import Message, MessageTypes
from .core.internal_protocols import CommunicatorProtocol
from .json_utils.rpc_generator import RPCGenerator
from .utils.message_handler import MessageHandler


class FakeContext:
    """A fake context instance, similar to the result of `zmq.Context.instance()."""

    def socket(self, socket_type):
        return FakeSocket(socket_type)

    def term(self):
        self.closed = True

    def destroy(self, linger=0):
        self.closed = True


class FakeSocket:
    """A fake socket mirroring zmq.socket API, useful for unit tests.

    :attr list _s: contains a list of messages sent via this socket.
    :attr list _r: List of messages which can be read.
    """

    def __init__(self, socket_type: int, *args) -> None:
        self.closed: bool = False

        # Added for testing purposes
        self.addr: Union[None, str] = None
        self.socket_type: int = socket_type
        # they contain a list of messages sent/received
        self._s: list[list[bytes]] = []
        self._r: list[list[bytes]] = []
        if socket_type == 2:  # zmq.SUB
            # empirical data shots, that you have to unsubscribe as many times as you have
            # subscribed, therefore a list is best
            self._subscriptions: list[bytes] = []

    def bind(self, addr: str) -> None:
        self.addr = addr

    def bind_to_random_port(self, addr: str, *args, **kwargs) -> int:
        self.addr = addr
        return 5

    def unbind(self, addr: Optional[str] = None) -> None:
        self.addr = None

    def connect(self, addr: str):
        self.addr = addr

    def disconnect(self, addr: Optional[str] = None) -> None:
        self.addr = None

    def poll(self, timeout: Optional[int] = None,
             flags: int = "PollEvent.POLLIN") -> int:  # type: ignore
        """Poll the socket for events.

        :returns: poll event mask (POLLIN, POLLOUT), 0 if the timeout was reached without an event.
        """
        return 1 if len(self._r) else 0

    def recv_multipart(self, flags: int = 0, *, copy: bool = True, track: bool = False
                       ) -> list[bytes]:
        return self._r.pop(0)

    def send_multipart(self, msg_parts: Sequence, flags: int = 0, copy: bool = True,
                       track: bool = False, **kwargs) -> None:
        for i, part in enumerate(msg_parts):
            if not isinstance(part, bytes):
                # Similar to real error message.
                raise TypeError(f"Frame {i} ({part}) does not support the buffer interface.")
        self._s.append(list(msg_parts))

    def subscribe(self, topic: Union[str, bytes]) -> None:
        if self.socket_type != 2:
            raise ValueError("Invalid argument")  # type is a ZMQError
        else:
            if isinstance(topic, str):
                topic = topic.encode()
            self._subscriptions.append(topic)

    def unsubscribe(self, topic: Union[str, bytes]) -> None:
        if self.socket_type != 2:
            raise ValueError("Invalid argument")  # type is a ZMQError
        else:
            if isinstance(topic, str):
                topic = topic.encode()
            try:
                self._subscriptions.remove(topic)
            except ValueError:
                pass  # not present

    def close(self, linger: Optional[int] = None) -> None:
        self.addr = None
        self.closed = True

    def set_hwm(self, hwm: int) -> None:
        self._hwm = hwm


class FakePoller:
    """A fake zmq poller."""
    def __init__(self) -> None:
        self._sockets: list[FakeSocket] = []

    def poll(self, timeout: Optional[int] = None) -> list[tuple[FakeSocket, Any]]:
        """Returns a list of events (socket, event_mask)"""
        events = []
        for sock in self._sockets:
            if sock.poll(timeout=timeout):
                events.append((sock, 1))
        return events

    def register(self, socket,
                 flags: int = "PollEvent.POLLIN",  # type: ignore
                 ) -> None:
        self._sockets.append(socket)

    def unregister(self, socket: FakeSocket) -> None:
        try:
            self._sockets.remove(socket)
        except ValueError:
            pass  # already removed


class FakeCommunicator(CommunicatorProtocol):
    """Contains lists with received (`_r`) and sent (`_s`) messages."""

    def __init__(self, name: str):
        super().__init__()
        self.name = name
        self.rpc_generator = RPCGenerator()
        self._r: list[Message] = []
        self._s: list[Message] = []

    def sign_in(self) -> None:
        self._signed_in = True

    def sign_out(self) -> None:
        self._signed_in = False

    def close(self) -> None:
        self._closed = True

    def send_message(self, message: Message) -> None:
        if not message.sender:
            message.sender = self.name.encode()
        self._s.append(message)

    def read_message(self, conversation_id: Optional[bytes] = None, timeout=None):
        return self._r.pop(0)

    def ask_message(self, message: Message, timeout=None) -> Message:
        self.send_message(message)
        return self.read_message(timeout=timeout)


class FakeDirector:
    """Supplements a regular director to create a fake one for testing.

    If you want to test the `SomeDirector` class, which directs `SomeActor`, you can create a
    subclass of the `FakeDirector` and `SomeDirector`.
    The newly created director will verify that methods are present in the `SomeActor` class,
    defined in the `remote_class` parameter.
    That way you can test, that `SomeDirector` calls an existing remote method.

    ..code::

        class FakeSomeDirector(FakeDirector, SomeDirector):
            pass

        @pytest.fixture
        def some_director():
            return FakeSomeDirector(remote_class=SomeActor)

        def test_some_method(some_director):
            # arrange
            some_director.return_value = 7  # define return value
            # act
            some_director.ask_rpc(method="method_name", arg1=5)  # returns 7
            # assert that the correct method name and kwargs are called
            assert some_director.method == "method_name"
            assert some_director.kwargs == {"arg1": 5}

    """

    return_value: Any  # value which the remote method should return
    method: str  # called method
    kwargs: dict[str, Any]  # kwargs sent to the method

    def __init__(self, remote_class, **kwargs):
        kwargs.setdefault("communicator", FakeCommunicator("communicator"))
        super().__init__(**kwargs)
        self.remote_class = remote_class

    def ask_rpc(
        self,
        method: str,
        actor: Optional[Union[bytes, str]] = None,
        additional_payload: Optional[Iterable[bytes]] = None,
        extract_additional_payload: bool = False,
        **kwargs,
    ) -> Any:
        assert hasattr(self.remote_class, method), f"Remote class does not have method '{method}'."
        self.method = method
        self.kwargs = kwargs
        return self.return_value

    def ask_rpc_async(
        self,
        method: str,
        actor: Optional[Union[bytes, str]] = None,
        additional_payload: Optional[Iterable[bytes]] = None,
        **kwargs,
    ) -> bytes:
        assert hasattr(self.remote_class, method), f"Remote class does not have method '{method}'."
        self.method = method
        self.kwargs = kwargs
        return b"conversation_id;"


def handle_request_message(handler: MessageHandler, method: str, *args, **kwargs) -> None:
    request = handler.rpc_generator.build_request_str(method, *args, **kwargs)
    handler.handle_message(Message("Handler", "Sender", request, message_type=MessageTypes.JSON))


def assert_response_is_result(handler: MessageHandler) -> Any:
    """Assert that response is result, not error and return the result."""
    response_message = Message.from_frames(*handler.socket._s[0])  # type: ignore
    result = handler.rpc_generator.get_result_from_response(response_message.payload[0])
    # get_result_from_response will raise an Exception if the response is an error.
    return result



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
import pickle
from typing import Any, Iterable, Optional, Union

import zmq

from ..core import PROXY_RECEIVING_PORT
from ..core.data_message import DataMessage, MessageTypes


class DataPublisher:
    """
    Publishing data via the LECO data protocol.

    :param str full_name: Name of the publishing Component
    :param str address: Address of the server, default is localhost.
    :param int port: Port of the server, defaults to 11100, default proxy.
    :param log: Logger to log to.

    Sending :class:`DataMessage` via the data protocol.

    Quantities may be expressed as a (magnitude number, units str) tuple.
    """

    full_name: str

    def __init__(
        self,
        full_name: str,
        host: str = "localhost",
        port: int = PROXY_RECEIVING_PORT,
        log: Optional[logging.Logger] = None,
        context: Optional[zmq.Context] = None,
        **kwargs,
    ) -> None:
        if log is None:
            self.log = logging.getLogger(f"{__name__}.Publisher")
        else:
            self.log = log.getChild("Publisher")
        self.log.info(f"Publisher started at {host}:{port}.")
        context = context or zmq.Context.instance()
        self.socket: zmq.Socket = context.socket(zmq.PUB)
        self.socket.connect(f"tcp://{host}:{port}")
        self.full_name = full_name
        super().__init__(**kwargs)

    def __del__(self) -> None:
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()

    def close(self) -> None:
        self.socket.close(1)

    def __call__(self, data: Any) -> None:
        """Publish `data`."""
        self.send_data(data=data)

    def send_message(self, message: DataMessage) -> None:
        """Send a data protocol message."""
        self.socket.send_multipart(message.to_frames())

    def send_data(
        self,
        data: Any,
        topic: Optional[Union[bytes, str]] = None,
        conversation_id: Optional[bytes] = None,
        message_type: Union[MessageTypes, int] = MessageTypes.NOT_DEFINED,
        additional_payload: Optional[Iterable[bytes]] = None,
    ) -> None:
        """Send the `data` via the data protocol."""
        message = DataMessage(
            topic=topic or self.full_name,
            data=data,
            conversation_id=conversation_id,
            message_type=message_type,
            additional_payload=additional_payload,
        )
        self.send_message(message)

    def send_legacy(self, data: dict[str, Any]) -> None:
        for key, value in data.items():
            # 234 is message type for legacy pickle: publish variable name as topic and pickle it
            self.send_data(topic=key, data=pickle.dumps(value), message_type=234)

    def set_full_name(self, full_name: str) -> None:
        """Set the full name of the data publisher.

        This method is useful for the listener's handler. That way a change of the listener's
        name or namespace is transferred easily to the publisher as well.

        .. code::

            listener = Listener()
            publisher = data_publisher(full_name=listener.full_name)
            listener.message_handler.register_on_name_change_method(publisher.set_full_name)

        """
        self.full_name = full_name



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from time import perf_counter
from typing import Optional, Protocol

import zmq

from ..core.internal_protocols import CommunicatorProtocol
from ..core.message import Message, MessageTypes
from ..json_utils.errors import JSONRPCError, DUPLICATE_NAME, NOT_SIGNED_IN


NOT_SIGNED_IN_ERROR_CODE = str(NOT_SIGNED_IN.code).encode()


class MessageBuffer:
    _messages: list[Message]
    _requested_ids: set[bytes]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._messages = []
        self._requested_ids = set()

    def add_conversation_id(self, conversation_id: bytes) -> None:
        """Add a conversation_id such that its response has to be recalled by name."""
        self._requested_ids.add(conversation_id)

    def remove_conversation_id(self, conversation_id: bytes) -> None:
        """Remove a conversation_id from the requested ids."""
        self._requested_ids.discard(conversation_id)

    def is_conversation_id_requested(self, conversation_id: bytes) -> bool:
        """Check whether this conversation_id is requested by someone."""
        return conversation_id in self._requested_ids

    def add_message(self, message: Message):
        """Add a message to the buffer."""
        self._messages.append(message)

    def retrieve_message(self, conversation_id: Optional[bytes] = None) -> Optional[Message]:
        """Retrieve the requested message or the next free one for `conversation_id=None`."""
        for i, msg in enumerate(self._messages):
            cid = msg.conversation_id
            if conversation_id == cid:
                self._requested_ids.discard(cid)
                return self._messages.pop(i)
            elif cid not in self._requested_ids and conversation_id is None:
                return self._messages.pop(i)
        return None

    def __len__(self):
        return len(self._messages)


class BaseCommunicator(CommunicatorProtocol, Protocol):
    """Abstract class of a Communicator with some logic.
    """

    socket: zmq.Socket
    log: logging.Logger
    namespace: Optional[str]
    message_buffer: MessageBuffer

    # Setup methods for call in init
    def setup_message_buffer(self) -> None:
        """Create the message buffer variables."""
        self.message_buffer = MessageBuffer()

    def close(self) -> None:
        """Close the connection."""
        self.socket.close(1)

    # Context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()

    # Base communication
    def _send_socket_message(self, message: Message) -> None:
        self.socket.send_multipart(message.to_frames())

    def send_message(self, message: Message) -> None:
        """Send a message, supplying sender information."""
        if not message.sender:
            message.sender = self.full_name.encode()
        self.log.debug(f"Sending {message}")
        self._send_socket_message(message=message)

    def sign_in(self) -> None:
        string = self.rpc_generator.build_request_str(method="sign_in")
        try:
            msg = self.ask(b"COORDINATOR", data=string, message_type=MessageTypes.JSON)
            self.interpret_rpc_response(msg)
        except JSONRPCError as exc:
            json_error = exc.rpc_error
            if json_error.code == DUPLICATE_NAME.code:
                self.log.warning("Sign in failed, the name is already used.")
            else:
                self.log.warning(f"Sign in failed, unknown error '{json_error}'.")
        except TimeoutError:
            self.log.error("Signing in timed out.")
        else:
            self.finish_sign_in(msg)

    def finish_sign_in(self, response_message: Message) -> None:
        self.namespace = response_message.sender_elements.namespace.decode()
        self.log.info(f"Signed in to Node '{self.namespace}'.")

    def heartbeat(self) -> None:
        """Send a heartbeat to the router."""
        self.log.debug("heartbeat")
        self.send_message(Message(b"COORDINATOR"))

    def sign_out(self) -> None:
        try:
            self.ask_rpc(b"COORDINATOR", method="sign_out")
        except TimeoutError:
            self.log.warning("Waiting for sign out response timed out.")
        except Exception as exc:
            self.log.exception("Signing out failed.", exc_info=exc)
        else:
            self.finish_sign_out()

    def finish_sign_out(self) -> None:
        self.log.info(f"Signed out from Node '{self.namespace}'.")
        self.namespace = None

    # Reading messages with buffer
    def _read_socket_message(self, timeout: Optional[float] = None) -> Message:
        """Read the next message from the socket, without further processing."""
        if self.socket.poll(int((timeout if timeout is not None else self.timeout) * 1000)):
            return Message.from_frames(*self.socket.recv_multipart())
        raise TimeoutError("Reading timed out")

    def _find_socket_message(self, conversation_id: Optional[bytes] = None,
                             timeout: Optional[float] = None,
                             ) -> Message:
        """Find a specific message among socket messages, storing the other ones in the buffer.

        :param conversation_id: Conversation ID to filter for, or next free message if None.
        """
        stop = perf_counter() + (timeout if timeout is not None else self.timeout)
        while True:
            msg = self._read_socket_message(timeout)
            self.check_for_not_signed_in_error(message=msg)
            cid = msg.conversation_id
            if conversation_id == cid:
                self.message_buffer.remove_conversation_id(conversation_id=cid)
                return msg
            elif self.message_buffer.is_conversation_id_requested(conversation_id=cid):
                self.message_buffer.add_message(msg)
            elif conversation_id is None:
                return msg
            else:
                self.message_buffer.add_message(msg)
            if perf_counter() > stop:
                # inside the loop to do it at least once, even if timeout is 0
                break
        raise TimeoutError("Message not found.")

    def check_for_not_signed_in_error(self, message: Message) -> None:
        if (message.sender_elements.name == b"COORDINATOR"
                and message.payload
                and b"error" in message.payload[0]
                and NOT_SIGNED_IN_ERROR_CODE in message.payload[0]):
            self.handle_not_signed_in()

    def read_message(self, conversation_id: Optional[bytes] = None,
                     timeout: Optional[float] = None) -> Message:
        message = self.message_buffer.retrieve_message(conversation_id=conversation_id)
        if message is None:
            message = self._find_socket_message(conversation_id=conversation_id, timeout=timeout)
        return message

    def handle_not_signed_in(self) -> None:
        self.namespace = None
        self.sign_in()
        self.log.warning("I was not signed in, signing in.")

    def ask_message(self, message: Message, timeout: Optional[float] = None) -> Message:
        self.send_message(message=message)
        return self.read_message(conversation_id=message.conversation_id, timeout=timeout)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.json_utils.errors import JSONRPCError
from pyleco.utils.message_handler import MessageHandler
from pyleco.test import (
    FakeCommunicator,
    FakePoller,
    FakeSocket,
    assert_response_is_result,
    FakeContext,
    handle_request_message,
)


@pytest.fixture
def poller() -> FakePoller:
    return FakePoller()


@pytest.fixture
def socket() -> FakeSocket:
    return FakeSocket(1)


@pytest.fixture
def sub_socket() -> FakeSocket:
    return FakeSocket(2)


def test_socket_unbind(socket: FakeSocket):
    socket.bind("abc")
    socket.unbind()
    assert socket.addr is None


def test_socket_disconnect(socket: FakeSocket):
    socket.connect("abc")
    socket.disconnect()
    assert socket.addr is None


@pytest.mark.parametrize("topic", ("string", b"bytes"))
def test_socket_subscribe(sub_socket: FakeSocket, topic):
    sub_socket.subscribe(topic)
    assert isinstance(sub_socket._subscriptions[-1], bytes)


def test_subscribe_fails_for_not_SUB(socket: FakeSocket):
    with pytest.raises(ValueError):
        socket.subscribe("abc")


@pytest.mark.parametrize("topic", ("topic", b"topic"))
def test_socket_unsubscribe(sub_socket: FakeSocket, topic):
    sub_socket._subscriptions.append(b"topic")
    sub_socket.unsubscribe(topic)
    assert b"topic" not in sub_socket._subscriptions


def test_unsubscribe_fails_for_not_SUB(socket: FakeSocket):
    with pytest.raises(ValueError):
        socket.unsubscribe("abc")


def test_socket_hwm(socket: FakeSocket):
    socket.set_hwm(5)
    assert socket._hwm == 5


class Test_FakePoller_unregister:
    def test_no_error_at_missing(self, poller: FakePoller):
        poller.unregister(FakeSocket(1))
        # assert no error is raised

    def test_unregister_removes_socket(self, poller: FakePoller):
        socket = FakeSocket(1)
        poller._sockets = [1, 2, socket, 4, 5]  # type: ignore
        poller.unregister(socket)
        assert socket not in poller._sockets


def test_FakeCommunicator_sign_in():
    fc = FakeCommunicator("")
    fc.sign_in()
    assert fc._signed_in is True


def test_assert_response_is_result_raises_exception_on_error():
    handler = MessageHandler("test", context=FakeContext())  # type: ignore
    handle_request_message(handler, "some_non_existing_method")
    with pytest.raises(JSONRPCError):
        assert_response_is_result(handler)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
import time

from unittest.mock import MagicMock

import pytest

from pyleco.core.message import Message
from pyleco.actors.locking_actor import LockingActor, AccessDeniedError
from pyleco.core.leco_protocols import LockingActorProtocol


class FantasyChannel:
    def __init__(self) -> None:
        self._prop = -1

    @property
    def channel_property(self):
        return self._prop

    @channel_property.setter
    def channel_property(self, value):
        self._prop = value

    @property
    def l_c_prop(self):
        return self._prop

    @l_c_prop.setter
    def l_c_prop(self, value):
        self._prop = value

    def channel_method(self, value):
        return 2 * value


class FantasyInstrument:
    """Some instrument to be controlled.

    The prefix "l" indicates properties etc. which should be locked.
    """
    def __init__(self, adapter, name="FantasyInstrument", *args, **kwargs):
        self.name = name
        self.adapter = adapter
        super().__init__()
        self.l_channel = FantasyChannel()
        self.l_channel.trace = FantasyChannel()  # type: ignore
        self.o_channel = FantasyChannel()
        self._l_prop = 5
        self._o_prop = 6

    @property
    def l_prop(self):
        return self._l_prop

    @l_prop.setter
    def l_prop(self, value):
        self._l_prop = value

    @property
    def o_prop(self):
        return self._o_prop

    @o_prop.setter
    def o_prop(self, value):
        self._o_prop = value

    def l_method(self, value):
        self._method_value = value

    def o_method(self, value):
        return value**2

    # methods for Instrument simulation
    def connect(self, *args):
        pass

    def disconnect(self, *args):
        pass


class FakeActor(LockingActor):
    def queue_readout(self):
        logging.getLogger().info(f"queue: {time.perf_counter()}")
        super().queue_readout()

    def heartbeat(self):
        logging.getLogger().info("beating")
        super().heartbeat()


@pytest.fixture()
def actor() -> FakeActor:
    actor = FakeActor(
        "test",
        FantasyInstrument,
        auto_connect={"adapter": MagicMock()},
        port=1234,
        protocol="inproc",
    )
    actor.next_beat = float("inf")
    return actor


resources = (
    "l_prop",  # a property
    "l_method",  # a method
    "l_channel",  # a channel
    "l_channel.l_c_prop",  # property of a channel
    "o_channel.l_c_prop",  # property of a channel
)


@pytest.fixture
def locked_actor(actor: LockingActor) -> LockingActor:
    actor.current_message = Message("rec", "owner")
    for r in resources:
        actor.lock(r)
    actor.current_message = Message("rec", "requester")
    return actor


class TestProtocolImplemented:
    protocol_methods = [m for m in dir(LockingActorProtocol) if not m.startswith("_")]

    def static_test_methods_are_present(self):
        def testing(component: LockingActorProtocol):
            pass

        testing(FakeActor(name="test", device_class=FantasyInstrument))

    @pytest.fixture
    def component_methods(self, actor: LockingActor):
        response = actor.rpc.process_request(
            '{"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}'
        )
        result = actor.rpc_generator.get_result_from_response(response)  # type: ignore
        return result.get("methods")

    @pytest.mark.parametrize("method", protocol_methods)
    def test_method_is_available(self, component_methods, method):
        for m in component_methods:
            if m.get("name") == method:
                return
        raise AssertionError(f"Method {method} is not available.")

class Test_check_access_rights:
    @pytest.mark.parametrize("resource", resources)
    def test_owner(self, locked_actor: LockingActor, resource):
        locked_actor.current_message = Message("rec", "owner")
        assert locked_actor.check_access_rights(resource) is True

    @pytest.mark.parametrize("resource", ("l_channel.channel_property", "l_channel.trace"))
    def test_owner_of_parent(self, locked_actor: LockingActor, resource):
        locked_actor.current_message = Message("rec", "owner")
        assert locked_actor.check_access_rights(resource) is True

    @pytest.mark.parametrize("resource", (None, *resources))
    def test_owner_of_device(self, actor: LockingActor, resource):
        """Only the device itself is locked, test access to parts."""
        actor.current_message = Message("rec", "owner")
        actor.lock(None)
        # act and assert
        assert actor.check_access_rights(resource) is True

    @pytest.mark.parametrize(
        "resource", (None, "o_prop", "o_method", "o_channel", "o_channel.channel_property")
    )
    def test_requester_True(self, locked_actor: LockingActor, resource):
        """Test that another requester may access unlocked resources."""
        locked_actor.current_message = Message("rec", "requester")
        assert locked_actor.check_access_rights(resource) is True

    @pytest.mark.parametrize(
        "resource", ("l_channel", "l_channel.channel_property", "l_prop", "o_channel.l_c_prop")
    )
    def test_requester_False(self, locked_actor: LockingActor, resource):
        # arrange
        locked_actor.force_unlock(None)
        # act
        locked_actor.current_message = Message("rec", "requester")
        assert locked_actor.check_access_rights(resource) is False

    @pytest.mark.parametrize("resource", (None, *resources))
    def test_not_owner_of_device(self, actor: LockingActor, resource):
        """Only the device itself is locked, test access to parts of it."""
        actor.current_message = Message("rec", "owner")
        actor.lock(None)
        actor.current_message = Message("rec", "requester")
        # act and assert
        assert actor.check_access_rights(resource) is True


@pytest.mark.parametrize(
    "resource",
    resources,
)
def test_lock_unlocked(actor: LockingActor, resource):
    actor.current_message = Message("rec", "owner")
    assert actor.lock(resource) is True
    assert actor._locks[resource] == b"owner"


@pytest.mark.parametrize(
    "resource",
    resources,
)
def test_lock_already_locked(locked_actor: LockingActor, resource):
    locked_actor.current_message = Message("rec", "owner")
    assert locked_actor.lock(resource) is True
    assert locked_actor._locks[resource] == b"owner"


@pytest.mark.parametrize(
    "resource",
    resources,
)
def test_lock_fail_as_already_locked(locked_actor: LockingActor, resource):
    locked_actor.current_message = Message("rec", "requester")
    assert locked_actor.lock(resource) is False
    assert locked_actor._locks[resource] == b"owner"


@pytest.mark.parametrize(
    "resource",
    resources,
)
def test_unlock_locked(locked_actor: LockingActor, resource):
    locked_actor.current_message = Message("rec", "owner")
    locked_actor.unlock(resource)
    assert resource not in locked_actor._locks


@pytest.mark.parametrize("resource", (None, "prop"))
def test_unlock_already_unlocked(actor: LockingActor, resource):
    actor.current_message = Message("rec", "requester")
    actor.unlock(resource)
    # assert no error is raised


@pytest.mark.parametrize(
    "resource",
    resources,
)
def test_unlock_fail_as_different_user(locked_actor: LockingActor, resource):
    locked_actor.current_message = Message("rec", "requester")
    # with pytest.raises(AccessDeniedError, match=resource):
    locked_actor.unlock(resource)
    assert locked_actor._locks[resource] == b"owner"


@pytest.mark.parametrize(
    "resource",
    ("l_channel.channel_method", "l_channel.trace"),
)
def test_lock_fail_for_child_of_locked_resource(locked_actor: LockingActor, resource):
    """If the parent is locked (e.g. the device), no child may be locked."""
    locked_actor.current_message = Message("rec", "requester")
    assert locked_actor.lock(resource) is False


# test device access
def test_get_parameters_successfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "owner")
    locked_actor.get_parameters(["l_prop", "l_channel.channel_property", "o_prop"])
    # assert that no error is raised


def test_get_parameters_unsuccessfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "requester")
    with pytest.raises(AccessDeniedError, match="'l_prop'"):
        locked_actor.get_parameters(["o_prop", "l_prop"])


def test_set_parameters_successfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "owner")
    locked_actor.set_parameters({"l_prop": 5, "l_channel.channel_property": 6})
    # assert that no error is raised


def test_set_parameters_unsuccessfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "requester")
    with pytest.raises(AccessDeniedError, match="'l_prop'"):
        locked_actor.set_parameters({"o_prop": 5, "l_prop": 6})


def test_call_action_successfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "owner")
    locked_actor.call_action("l_method", [5])
    # assert that no error is raised


def test_call_action_unsuccessfully(locked_actor: LockingActor):
    locked_actor.current_message = Message("rec", "requester")
    with pytest.raises(AccessDeniedError, match="'l_method'"):
        locked_actor.call_action("l_method", [5])



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from typing import Any, Iterable, Optional, Sequence, Union

from ..core.internal_protocols import CommunicatorProtocol
from ..utils.communicator import Communicator
from ..utils.log_levels import get_leco_log_level
from ..core.serialization import generate_conversation_id
from ..core.message import Message, MessageTypes


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Director:
    """Basic director handling.

    They can be used as a ContextManager:
    .. code::

        with BaseDirector() as d:
            d.get_properties(["property1", "property2"])

    :param actor: Default name of the Actor to communicate with. Stored as :attr:`actor`.
    :param communicator: A Communicator class to communicate with the actor.
        If None, create a new Communicator instance.
    :param name: The name of this Director.
    """

    def __init__(self, actor: Optional[Union[bytes, str]] = None,
                 communicator: Optional[CommunicatorProtocol] = None,
                 name: str = "Director",
                 **kwargs) -> None:
        self.actor = actor
        if communicator is None:
            communicator = Communicator(name=name, **kwargs)
            try:
                communicator.sign_in()
            except TimeoutError:
                log.error("Signing in timed out!")
            kwargs = {}
            self._own_communicator = True  # whether to sign out or not.
        else:
            self._own_communicator = False
        self.communicator = communicator
        self.generator = communicator.rpc_generator
        super().__init__(**kwargs)

    def close(self) -> None:
        if self._own_communicator:
            self.communicator.close()

    def sign_out(self) -> None:
        """Sign the communicator out."""
        self.communicator.sign_out()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close()

    # Message handling
    def ask_message(self, actor: Optional[Union[bytes, str]] = None,
                    data: Optional[Any] = None, **kwargs) -> Message:
        actor = self._actor_check(actor)
        log.debug(f"Asking {actor!r} with message '{data}'.")
        response = self.communicator.ask(actor, data=data, **kwargs)
        log.debug(f"Data '{response.data}' received.")
        return response

    def _actor_check(self, actor: Optional[Union[bytes, str]]) -> Union[bytes, str]:
        actor = actor or self.actor
        if actor is None:
            raise ValueError("Some actor has to be specified.")
        return actor

    # Helper methods
    def _prepare_call_action_params(self, args: tuple[Any, ...],
                                    kwargs: dict[str, Any]) -> dict[str, Any]:
        """Generate a params dictionary for the call action method."""
        params: dict[str, Any] = {}
        if args:
            params["args"] = args
        if kwargs:
            params["kwargs"] = kwargs
        return params

    # Remote control synced
    def ask_rpc(
        self,
        method: str,
        actor: Optional[Union[bytes, str]] = None,
        additional_payload: Optional[Iterable[bytes]] = None,
        extract_additional_payload: bool = False,
        **kwargs,
    ) -> Any:
        """Remotely call the `method` procedure on the `actor` and return the return value."""
        receiver = self._actor_check(actor)
        return self.communicator.ask_rpc(
            receiver=receiver,
            method=method,
            additional_payload=additional_payload,
            extract_additional_payload=extract_additional_payload,
            **kwargs,
        )

    #   Component
    def get_rpc_capabilities(self, actor: Optional[Union[bytes, str]] = None) -> dict:
        """Get a list of the remotely callable procedures of the actor."""
        return self.ask_rpc(method="rpc.discover", actor=actor)

    def shut_down_actor(self, actor: Optional[Union[bytes, str]] = None) -> None:
        """Stop the actor."""
        self.ask_rpc(method="shut_down", actor=actor)

    def set_actor_log_level(self, level: Union[str, int], actor: Optional[Union[bytes, str]] = None
                            ) -> None:
        """Set the log level of the actor."""
        if isinstance(level, int):
            level = get_leco_log_level(level).value
        self.ask_rpc("set_log_level", level=level, actor=actor)

    #   Actor
    def get_parameters(self, parameters: Union[str, Sequence[str]],
                       actor: Optional[Union[bytes, str]] = None) -> dict[str, Any]:
        """Get the values of these `properties` (list, tuple)."""
        if isinstance(parameters, str):
            parameters = (parameters,)
        response = self.ask_rpc(method="get_parameters", parameters=parameters, actor=actor)
        if not isinstance(response, dict):
            raise ConnectionError(f"{response} returned, but dict expected.")
        return response

    def set_parameters(self, parameters: dict[str, Any],
                       actor: Optional[Union[bytes, str]] = None) -> None:
        """Set the `properties` dictionary."""
        self.ask_rpc(method="set_parameters", parameters=parameters, actor=actor)

    def call_action(self, action: str, *args, actor: Optional[Union[bytes, str]] = None,
                    **kwargs) -> Any:
        """Call an action remotely and return its return value.

        :param str action: Name of the action to call. If you have positional arguments, this
            parameter has to be the first positional argument
        :param \\*args: Arguments for the action to call.
        :param str actor: Name of the actor to execute the action.
            Defaults to the stored actor name.
        :param \\**kwargs: Keyword arguments for the action to call.
        """
        params = self._prepare_call_action_params(args, kwargs)
        return self.ask_rpc("call_action", action=action, actor=actor, **params)

    # Async methods: Just send, read later.
    def send(
        self,
        actor: Optional[Union[bytes, str]] = None,
        data=None,
        additional_payload: Optional[Iterable[bytes]] = None,
        **kwargs,
    ) -> bytes:
        """Send a request and return the conversation_id."""
        actor = self._actor_check(actor)
        cid0 = generate_conversation_id()
        self.communicator.send(
            actor, conversation_id=cid0, data=data, additional_payload=additional_payload, **kwargs
        )
        return cid0

    def ask_rpc_async(
        self,
        method: str,
        actor: Optional[Union[bytes, str]] = None,
        additional_payload: Optional[Iterable[bytes]] = None,
        **kwargs,
    ) -> bytes:
        """Send a rpc request, the response can be read later with :meth:`read_rpc_response`."""
        string = self.generator.build_request_str(method=method, **kwargs)
        return self.send(
            actor=actor,
            data=string,
            message_type=MessageTypes.JSON,
            additional_payload=additional_payload,
        )

    def read_rpc_response(
        self,
        conversation_id: Optional[bytes] = None,
        extract_additional_payload: bool = False,
        **kwargs,
    ) -> Any:
        """Read the response value corresponding to a request with a certain `conversation_id`."""
        response_message = self.communicator.read_message(conversation_id=conversation_id, **kwargs)
        return self.communicator.interpret_rpc_response(
            response_message=response_message, extract_additional_payload=extract_additional_payload
        )

    #   Actor
    def get_parameters_async(self, parameters: Union[str, Sequence[str]],
                             actor: Optional[Union[bytes, str]] = None) -> bytes:
        """Request the values of these `properties` (list, tuple) and return the conversation_id.

        You can use :meth:`read_rpc_response` to read the response.
        """
        if isinstance(parameters, str):
            parameters = (parameters,)
        # return self.send(data=[[Commands.GET, properties]])
        return self.ask_rpc_async(method="get_parameters", parameters=parameters, actor=actor)

    def set_parameters_async(self, parameters: dict[str, Any],
                             actor: Optional[Union[bytes, str]] = None) -> bytes:
        """Set the `properties` dictionary and return the conversation_id.

        You can use :meth:`read_rpc_response` to read the response.
        """
        # return self.send(data=[[Commands.SET, properties]])
        return self.ask_rpc_async(method="set_parameters", parameters=parameters, actor=actor)

    def call_action_async(self, action: str, *args, actor: Optional[Union[bytes, str]] = None,
                          **kwargs) -> bytes:
        """Call a method remotely and return the conversation_id.

        You can use :meth:`read_rpc_response` to read the response.

        :param str action: Name of the action to call.
        :param \\*args: Arguments for the action to call.
        :param str actor: Name of the actor to execute the action.
            Defaults to the stored actor name.
        :param \\**kwargs: Keyword arguments for the action to call.
        """
        params = self._prepare_call_action_params(args, kwargs)
        return self.ask_rpc_async(method="call_action", action=action, actor=actor, **params)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging

import pytest

from pyleco.utils.log_levels import get_leco_log_level, LogLevels, PythonLogLevels


@pytest.mark.parametrize("level, value", (
    (logging.DEBUG, LogLevels.DEBUG),
    (logging.INFO, LogLevels.INFO),
    (logging.WARNING, LogLevels.WARNING),
    (logging.ERROR, LogLevels.ERROR),
    (logging.CRITICAL, LogLevels.CRITICAL),
))
def test_get_leco_log_level(level, value):
    assert get_leco_log_level(level) == value


def test_failing_get_leco_log_level():
    with pytest.raises(ValueError):
        get_leco_log_level(5)


def test_PythonLogLevels():
    assert PythonLogLevels["DEBUG"] == logging.DEBUG



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
from socket import gethostname
from time import sleep
import threading

import pytest

from pyleco.core.message import Message, MessageTypes
from pyleco.utils.listener import Listener
from pyleco.utils.communicator import Communicator

# Test the Coordinator and its Director in a live test
from pyleco.directors.coordinator_director import CoordinatorDirector
from pyleco.coordinators.coordinator import Coordinator


# Constants
PORT = 60001
PORT2 = PORT + 1
PORT3 = PORT + 2
TALKING_TIME = 0.5  # s
TIMEOUT = 2  # s


hostname = gethostname()
testlevel = 30
# pytest.skip("Takes too long.", allow_module_level=True)


def start_coordinator(namespace: str, port: int, coordinators=None, stop_event=None, **kwargs):
    with Coordinator(namespace=namespace, port=port, **kwargs) as coordinator:
        coordinator.routing(coordinators=coordinators, stop_event=stop_event)


@pytest.fixture(scope="module")
def leco():
    """A leco setup."""
    glog = logging.getLogger()
    glog.setLevel(logging.DEBUG)
    # glog.addHandler(logging.StreamHandler())
    log = logging.getLogger("test")
    threads = []
    stop_events = [threading.Event(), threading.Event(), threading.Event()]
    threads.append(threading.Thread(target=start_coordinator,
                                    kwargs=dict(namespace="N1", port=PORT,
                                                stop_event=stop_events[0])))
    threads.append(threading.Thread(target=start_coordinator,
                                    kwargs=dict(namespace="N2", port=PORT2,
                                                stop_event=stop_events[1])))
    threads.append(threading.Thread(target=start_coordinator,
                                    kwargs=dict(namespace="N3", port=PORT3,
                                                stop_event=stop_events[2])))
    for thread in threads:
        thread.daemon = True
        thread.start()
    listener = Listener(name="Controller", port=PORT, timeout=TIMEOUT)
    listener.start_listen()
    sleep(TALKING_TIME)  # time for setup
    yield listener.get_communicator()
    log.info("Tearing down")
    listener.stop_listen()
    for event in stop_events:
        event.set()
    for thread in threads:
        thread.join(0.5)


@pytest.mark.skipif(testlevel < 0, reason="reduce load")
def test_startup(leco: Communicator):
    with CoordinatorDirector(communicator=leco) as d:
        assert d.get_local_components() == ["Controller"]
        assert d.get_nodes() == {"N1": f"{hostname}:{PORT}"}


@pytest.mark.skipif(testlevel < 1, reason="reduce load")
def test_connect_N1_to_N2(leco: Communicator):
    with CoordinatorDirector(communicator=leco) as d:
        d.add_nodes({"N2": f"localhost:{PORT2}"})
        sleep(TALKING_TIME)  # time for coordinators to talk
        # assert that the N1.COORDINATOR knows about N2
        assert d.get_nodes() == {"N1": f"{hostname}:{PORT}", "N2": f"localhost:{PORT2}"}
        # assert that the listener can contact N2.COORDINATOR
        assert d.ask_rpc(actor="N2.COORDINATOR", method="pong") is None


@pytest.mark.skipif(testlevel < 2, reason="reduce load")
def test_Component_to_Component_via_1_Coordinator(leco: Communicator):
    with Communicator(name="whatever", port=PORT) as c:
        assert c.ask_rpc("N1.Controller", method="pong") is None


@pytest.mark.skipif(testlevel < 2, reason="reduce load")
def test_Component_to_Component_via_2_Coordinators(leco: Communicator):
    with Communicator(name="whatever", port=PORT2) as c:
        response = c.ask("N1.Controller", data={"id": 1, "method": "pong", "jsonrpc": "2.0"},
                         message_type=MessageTypes.JSON)
        assert response == Message(
            b'N2.whatever', b'N1.Controller', data={"id": 1, "result": None, "jsonrpc": "2.0"},
            header=response.header)


@pytest.mark.skipif(testlevel < 2, reason="reduce load")
def test_Component_lists_propgate_through_Coordinators(leco: Communicator):
    """Test that Component lists are propagated from one Coordinator to another."""
    with CoordinatorDirector(actor="N2.COORDINATOR", name="whatever", port=PORT2) as d:
        assert d.get_global_components() == {"N1": ["Controller"], "N2": ["whatever"]}


@pytest.mark.skipif(testlevel < 2, reason="reduce load")
def test_Component_to_second_coordinator(leco: Communicator):
    assert leco.ask_rpc("N2.COORDINATOR", method="pong") is None


def test_sign_in_rejected_for_duplicate_name(leco: Communicator):
    with pytest.raises(ConnectionRefusedError):
        with Communicator(name="Controller", port=PORT):
            pass


@pytest.mark.skipif(testlevel < 3, reason="reduce load")
def test_connect_N3_to_N2(leco: Communicator):
    with CoordinatorDirector(name="whatever", port=PORT3) as d1:
        d1.add_nodes({"N2": f"localhost:{PORT2}"})

    sleep(TALKING_TIME)  # time for coordinators to talk
    with CoordinatorDirector(actor="COORDINATOR", communicator=leco) as d2:
        assert d2.get_nodes() == {"N1": f"{hostname}:{PORT}", "N2": f"localhost:{PORT2}",
                                  "N3": f"{hostname}:{PORT3}"}


@pytest.mark.skipif(testlevel < 4, reason="reduce load")
def test_shutdown_N3(leco: Communicator):
    with CoordinatorDirector(actor="N3.COORDINATOR", name="whatever", port=PORT3) as d1:
        d1.shut_down_actor()

    sleep(TALKING_TIME)  # time for coordinators to talk
    with CoordinatorDirector(actor="COORDINATOR", communicator=leco) as d2:
        assert d2.get_nodes() == {"N1": f"{hostname}:{PORT}", "N2": f"localhost:{PORT2}"}



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from threading import Thread, Event
from time import sleep
from typing import Any, Callable, Optional, Union

from ..core import PROXY_SENDING_PORT, COORDINATOR_PORT
from .pipe_handler import PipeHandler, CommunicatorPipe

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Listener:
    """Listening on incoming messages in a separate thread.

    On one side it handles incoming messages (data and control protocol) in another thread.
    On the other side, it offers the :meth:`get_communicator` method, which returns a
    :class:`Communicator`, offering communication to the network.

    Call :meth:`.start_listen()` to actually listen.

    ..code::

        listener = Listener()
        listener.start_listen()  # starts a message handler in another thread
        communicator = listener.get_communicator()  # get a Communicator endpoint for this thread
        response = communicator.ask_message(some_message_object)

    :param name: Name to listen under for control commands.
    :param int data_port: Port number for the data protocol.
    :param logger: Logger instance whose logs should be published. Defaults to "__main__".
    """

    communicator: CommunicatorPipe
    message_handler: PipeHandler

    def __init__(self,
                 name: str,
                 host: str = "localhost",
                 port: int = COORDINATOR_PORT,
                 data_host: Optional[str] = None,
                 data_port: int = PROXY_SENDING_PORT,
                 logger: Optional[logging.Logger] = None,
                 timeout: float = 1,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        log.info(f"Start Listener for '{name}'.")

        self.name = name
        self.logger = logger
        self.timeout = timeout

        self.coordinator_address = host, port
        self.data_address = data_host or host, data_port

    def close(self) -> None:
        """Close everything."""
        self.stop_listen()

    @property
    def name(self) -> str:
        try:
            return self.communicator.name
        except AttributeError:
            return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        try:
            self.communicator.name = value
        except AttributeError:
            pass

    # Methods to control the Listener
    def start_listen(self) -> None:
        """Start to listen in a thread."""
        self.stop_listen()
        self.stop_event = Event()
        self.thread = Thread(
            target=self._listen,
            args=(
                self.name,
                self.stop_event,
                self.coordinator_address[0],
                self.coordinator_address[1],
                self.data_address[0],
                self.data_address[1],
            ))
        self.thread.daemon = True
        self.thread.start()
        for _ in range(10):
            sleep(0.05)
            try:
                self.communicator: CommunicatorPipe = self.message_handler.get_communicator(
                    timeout=self.timeout)
            except AttributeError:
                pass
            else:
                log.addHandler(self.message_handler.log_handler)
                if self.logger is not None:
                    self.logger.addHandler(self.message_handler.log_handler)
                return
        raise TimeoutError("PipeHandler has not started after 0.5 s.")

    def get_communicator(self, **kwargs) -> CommunicatorPipe:
        """Get the communicator for the calling thread, creating one if necessary."""
        kwargs.setdefault("timeout", self.timeout)
        return self.message_handler.get_communicator(**kwargs)

    def register_rpc_method(self, method: Callable[..., Any], **kwargs) -> None:
        """Register a method for calling with the current message handler.

        If you restart the listening, you have to register the method anew.
        """
        self.message_handler.register_rpc_method(method=method, **kwargs)

    def register_binary_rpc_method(
        self,
        method: Callable[..., Union[Any, tuple[Any, list[bytes]]]],
        accept_binary_input: bool = False,
        return_binary_output: bool = False,
        **kwargs,
    ) -> None:
        """Register a binary method for calling with the current message handler.

        If you restart the listening, you have to register the method anew.
        """
        self.message_handler.register_binary_rpc_method(
            method=method,
            accept_binary_input=accept_binary_input,
            return_binary_output=return_binary_output,
            **kwargs,
        )

    def stop_listen(self) -> None:
        """Stop the listener Thread."""
        try:
            if self.thread.is_alive():
                log.debug("Stopping listener thread.")
                self.stop_event.set()
                self.thread.join()
                self.message_handler.close()
                log.removeHandler(self.message_handler.log_handler)
                if self.logger is not None:
                    self.logger.removeHandler(self.message_handler.log_handler)
        except AttributeError:
            pass

    """
    Methods below are executed in the thread, DO NOT CALL DIRECTLY!
    """

    def _listen(self, name: str, stop_event: Event, coordinator_host: str, coordinator_port: int,
                data_host: str, data_port: int) -> None:
        """Start a PipeHandler, which has to be executed in a separate thread."""
        self.message_handler = PipeHandler(name, host=coordinator_host, port=coordinator_port,
                                           data_host=data_host, data_port=data_port)
        self.message_handler.listen(stop_event=stop_event)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
These classes show pyleco's internal API of tools that talk with the LECO protocol.

They are not defined by LECO itself as it does not touch the message transfer.

Any Component could use these tools in order to send and read messages.
For example a Director might use these tools to direct an Actor.
"""

from __future__ import annotations
from typing import Any, Optional, Protocol, Iterable, Union

from .message import Message, MessageTypes
from ..json_utils.rpc_generator import RPCGenerator


class CommunicatorProtocol(Protocol):
    """A helper class for a Component, to communicate via the LECO protocol.

    For example a Director might use such a class to send/read messages to/from an Actor.
    """

    name: str
    namespace: Optional[str] = None
    rpc_generator: RPCGenerator
    timeout: float = 1  # default reading timeout in seconds

    @property
    def full_name(self) -> str:
        return self.name if self.namespace is None else ".".join((self.namespace, self.name))

    def sign_in(self) -> None: ...  # pragma: no cover

    def sign_out(self) -> None: ...  # pragma: no cover

    def send_message(self, message: Message) -> None: ...  # pragma: no cover

    def read_message(
        self, conversation_id: Optional[bytes], timeout: Optional[float] = None
    ) -> Message: ...  # pragma: no cover

    def ask_message(
        self, message: Message, timeout: Optional[float] = None
    ) -> Message: ...  # pragma: no cover

    def close(self) -> None: ...  # pragma: no cover

    # Utilities
    def send(
        self,
        receiver: Union[bytes, str],
        conversation_id: Optional[bytes] = None,
        data: Optional[Any] = None,
        **kwargs,
    ) -> None:
        """Send a message based on kwargs."""
        self.send_message(
            message=Message(receiver=receiver, conversation_id=conversation_id, data=data, **kwargs)
        )

    def ask(
        self,
        receiver: Union[bytes, str],
        conversation_id: Optional[bytes] = None,
        data: Optional[Any] = None,
        timeout: Optional[float] = None,
        **kwargs,
    ) -> Message:
        """Send a message based on kwargs and retrieve the response."""
        return self.ask_message(
            message=Message(
                receiver=receiver, conversation_id=conversation_id, data=data, **kwargs
            ),
            timeout=timeout,
        )

    def interpret_rpc_response(
        self, response_message: Message, extract_additional_payload: bool = False
    ) -> Union[Any, tuple[Any, list[bytes]]]:
        """Retrieve the return value of a RPC response and optionally the additional payload."""
        result = self.rpc_generator.get_result_from_response(response_message.payload[0])
        if extract_additional_payload:
            return result, response_message.payload[1:]
        else:
            return result

    def ask_rpc(
        self,
        receiver: Union[bytes, str],
        method: str,
        timeout: Optional[float] = None,
        additional_payload: Optional[Iterable[bytes]] = None,
        extract_additional_payload: bool = False,
        **kwargs,
    ) -> Any:
        """Send a JSON-RPC request (with method \\**kwargs) and return the response value."""
        string = self.rpc_generator.build_request_str(method=method, **kwargs)
        response = self.ask(
            receiver=receiver,
            data=string,
            message_type=MessageTypes.JSON,
            additional_payload=additional_payload,
            timeout=timeout,
        )
        return self.interpret_rpc_response(
            response, extract_additional_payload=extract_additional_payload
        )


class SubscriberProtocol(Protocol):
    """A helper class to subscribe to data protocol topics."""

    def subscribe_single(self, topic: bytes) -> None: ...  # pragma: no cover

    def unsubscribe_single(self, topic: bytes) -> None: ...  # pragma: no cover

    def unsubscribe_all(self) -> None: ...  # pragma: no cover

    def subscribe(self, topics: Union[str, Iterable[str]]) -> None:
        """Subscribe to a topic or list of topics."""
        if isinstance(topics, str):
            self.subscribe_single(topics.encode())
        else:
            for topic in topics:
                self.subscribe_single(topic.encode())

    def unsubscribe(self, topics: Union[str, Iterable[str]]) -> None:
        """Unsubscribe to a topic or list of topics."""
        if isinstance(topics, str):
            self.unsubscribe_single(topics.encode())
        else:
            for topic in topics:
                self.unsubscribe_single(topic.encode())



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from typing import Union, Sequence

import datetime
try:
    from enum import StrEnum  # type: ignore
except ImportError:  # pragma: no cover
    # For python<3.11
    from enum import Enum

    class StrEnum(str, Enum):  # type: ignore
        pass
import json
import logging
from threading import Lock
from typing import Any, Callable, Optional, Iterable

try:
    import numpy as np  # type: ignore[import-not-found]
except ModuleNotFoundError:
    def average(values: Sequence[Union[float, int]]) -> float:
        return sum(values) / len(values)
else:
    average = np.average  # type: ignore

if __name__ == "__main__":  # pragma: no cover
    from pyleco.utils.timers import RepeatingTimer
    from pyleco.utils.extended_message_handler import ExtendedMessageHandler, DataMessage
    from pyleco.utils.parser import parser, parse_command_line_parameters
    from pyleco.utils.data_publisher import DataPublisher
else:
    from ..utils.timers import RepeatingTimer
    from ..utils.extended_message_handler import ExtendedMessageHandler, DataMessage
    from ..utils.parser import parser, parse_command_line_parameters
    from ..utils.data_publisher import DataPublisher


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
StrFormatter = logging.Formatter("%(asctime)s\t%(levelname)s\t%(name)s\t%(message)s")

nan = float("nan")


class TriggerTypes(StrEnum):
    TIMER = "timer"
    VARIABLE = "variable"
    NONE = "none"


class ValuingModes(StrEnum):
    LAST = "last"
    AVERAGE = "average"


class DataLogger(ExtendedMessageHandler):
    """Collect data and save it to the disk, if required.

    The data logger listens to commands via the control protocol and to data via the data protocol.
    Whenever triggered, either by a timer or by receiving certain data via data protocol, it
    generates a data point.
    Each new datapoint is published via the data protocol, such that a user might follow the data
    acquisition.
    The data point contains values for each variable. The value is either the last one received
    since the last data point, or the average of all values received sind the last data point, or
    `float("nan")`.

    If desired, the datalogger may save all datapoints to disk.

    .. code::

        datalogger = DataLogger()
        datalogger.listen()  # listen until a shutdown signal is received.
        # Now you may send a "start_collecting" message to start a measurement.
    """

    # TODO names
    tmp: dict[str, list[Any]]  # contains all values since last datapoint
    lists: dict[str, list[Any]]  # contains datapoints.
    units: dict[str, Any]  # contains the units of the variables  TODO TBD what the value is.
    last_datapoint: dict[str, Any]
    last_save_name: str = ""

    # configuration variables
    trigger_type: TriggerTypes = TriggerTypes.NONE
    _last_trigger_type: TriggerTypes = TriggerTypes.NONE
    trigger_timeout: float = 1
    trigger_variable: str = ""
    value_repeating: bool = False
    valuing: Callable[[list], Any]

    def __init__(self, name: str = "DataLoggerN", directory: str = ".", **kwargs) -> None:
        super().__init__(name=name, **kwargs)
        self.directory = directory
        self.publisher = DataPublisher(full_name=name)
        self.valuing = average
        # Initialize values
        self.list_lock = Lock()
        self.reset_data_storage()
        self.units = {}
        # TODO add auto_save functionality?

    def register_rpc_methods(self) -> None:
        super().register_rpc_methods()
        self.register_rpc_method(self.set_valuing_mode)  # TODO offer during a measurement?
        self.register_rpc_method(self.start_collecting)
        self.register_rpc_method(self.save_data)
        self.register_rpc_method(self.stop_collecting)
        self.register_rpc_method(self.get_last_datapoint)
        self.register_rpc_method(self.get_list_length)
        self.register_rpc_method(self.get_last_save_name)
        self.register_rpc_method(self.get_configuration)

    def __del__(self) -> None:
        self.stop_collecting()

    def _listen_setup(self, start_data: Optional[dict[str, Any]] = None,  # type: ignore[override]
                      **kwargs):
        poller = super()._listen_setup(**kwargs)
        if start_data is not None:
            self.start_collecting(**start_data)
        return poller

    def _listen_close(self, waiting_time: Optional[int] = None) -> None:
        self.stop_collecting()
        super()._listen_close(waiting_time=waiting_time)

    def set_full_name(self, full_name: str) -> None:
        super().set_full_name(full_name=full_name)
        self.publisher.full_name = full_name

    # Data management
    def handle_subscription_message(self, message: DataMessage) -> None:
        sender = message.topic.decode()
        try:
            content: dict[str, Any] = message.data  # type: ignore
            modified_dict = {".".join((sender, k)): v for k, v in content.items()}
        except Exception:
            log.exception(f"Could not decode message {message}.")
        else:
            self.handle_subscription_data(modified_dict)

    def handle_subscription_data(self, data: dict[str, Any]) -> None:
        """Store `data` dict in `tmp`"""
        with self.list_lock:
            for key, value in data.items():
                try:
                    self.tmp[key].append(value)
                except KeyError:
                    log.debug("Got value for '%s', but no list present.", key)
        if self.trigger_type == TriggerTypes.VARIABLE and self.trigger_variable in data.keys():
            self.make_datapoint()

    def make_datapoint(self) -> dict[str, Any]:
        """Store a datapoint."""
        datapoint = self.calculate_data()
        self.last_datapoint = datapoint
        if self.namespace is not None:
            self.publisher.send_data(data=self.last_datapoint)
        return datapoint

    def calculate_data(self) -> dict[str, Any]:
        """Calculate data for a data point and return the data point."""
        datapoint = {}
        with self.list_lock:
            if 'time' in self.lists.keys():
                now = datetime.datetime.now(datetime.timezone.utc)
                today = datetime.datetime.combine(
                    self.today, datetime.time(), datetime.timezone.utc
                )
                time = (now - today).total_seconds()
                self.tmp['time'].append(time)
            for variable, datalist in self.lists.items():
                value = datapoint[variable] = self.calculate_single_data(
                    variable, self.tmp[variable]
                )
                datalist.append(value)
            for key in self.tmp.keys():
                self.tmp[key].clear()
            return datapoint

    def calculate_single_data(self, variable: str, tmp: list):
        if tmp:
            value = self.valuing(tmp)
        elif self.value_repeating:
            try:
                # no lock, as this method is called in in a locked environment!
                value = self.lists[variable][-1]
            except (KeyError, IndexError):  # No last value present.
                value = nan
        else:
            value = nan
        return value

    @staticmethod
    def last(data: list[Any]) -> Any:
        """Return the last value of an iterable with error handling."""
        try:
            return data[-1]
        except TypeError:
            return data
        except IndexError:
            # empty list
            return nan

    # Control
    def start_collecting(self, *,
                         variables: Optional[list[str]] = None,
                         units: Optional[dict[str, Any]] = None,
                         trigger_type: Optional[TriggerTypes] = None,  # TODO also str, but openrpc
                         trigger_timeout: Optional[float] = None,
                         trigger_variable: Optional[str] = None,
                         valuing_mode: Optional[ValuingModes] = None,  # TODO also str, but openrpc
                         value_repeating: Optional[bool] = None,
                         ) -> None:
        """Start collecting data.

        If you do not give a specific parameter, the value of the last measurement is used again.
        """
        self.stop_collecting()
        log.info(f"Start collecting data. Trigger: {trigger_type}, {trigger_timeout}, "
                 f"{trigger_variable}; subscriptions: {variables}")
        self.today = datetime.datetime.now(datetime.timezone.utc).date()
        self.trigger_type = TriggerTypes(trigger_type) if trigger_type else self._last_trigger_type
        self._last_trigger_type = self.trigger_type
        if trigger_timeout is not None:
            self.trigger_timeout = trigger_timeout
        if trigger_variable is not None:
            self.trigger_variable = trigger_variable
        if value_repeating is not None:
            self.value_repeating = value_repeating
        if self.trigger_type == TriggerTypes.TIMER:
            self.start_timer_trigger(timeout=self.trigger_timeout)
        self.set_valuing_mode(valuing_mode=valuing_mode)
        self.setup_variables(self.lists.keys() if variables is None else variables)
        self.units = units if units else {}

    def setup_variables(self, variables: Iterable[str]) -> None:
        """Subscribe to the variables."""
        self.reset_data_storage()
        subscriptions: set[str] = set()
        for variable in variables:
            if "." in variable:
                # this is the new style: topic is sender name, data is in content
                parts = variable.split(".")
                if len(parts) == 2:
                    # assume to be in the same namespace
                    if self.namespace is None:
                        log.error(f"Cannot subscribe to '{variable}' as the namespace is not known.")  # noqa
                        continue
                    parts.insert(0, self.namespace)
                    variable = ".".join(parts)
                subscriptions.add(".".join(parts[:2]))
            else:
                # old style: topic is variable name
                subscriptions.add(variable)
            with self.list_lock:
                self.lists[variable] = []
                self.tmp[variable] = []
        self.subscribe(topics=subscriptions)

    def reset_data_storage(self) -> None:
        """Reset the data storage."""
        with self.list_lock:
            self.tmp = {}
            self.lists = {}
        self.last_datapoint = {}

    def start_timer_trigger(self, timeout: float) -> None:
        self.timer = RepeatingTimer(timeout, self.make_datapoint)
        self.timer.start()

    def set_valuing_mode(self, valuing_mode: Optional[ValuingModes]) -> None:  # also str
        if valuing_mode == ValuingModes.LAST:
            self.valuing = self.last
        elif valuing_mode == ValuingModes.AVERAGE:
            self.valuing = average
        elif valuing_mode is None:
            pass  # already setup

    def save_data(self, meta: Optional[dict] = None, suffix: str = "", header: str = "") -> str:
        """Save the data.

        :param addr: Reply address for the filename.
        :param dict meta: The meta data to save. Use e.g. in subclass
            Protected keys: units, today, name, configuration, user.
        :param str suffix: Suffix to append to the filename.
        :return str: Name of the saved file.
        """
        # Preparation.
        if meta is None:
            meta = {}
        folder = self.directory
        # Pickle the header and lists.
        file_name = datetime.datetime.now().strftime("%Y_%m_%dT%H_%M_%S") + suffix
        meta.update({
            'units': self.units,
            'today': self.today.isoformat(),
            'file_name': file_name,
            'logger_name': self.full_name,
            'configuration': self.get_configuration(),
            # 'user': self.user_data,  # user stored meta data
        })
        try:
            with self.list_lock:
                with open(f"{folder}/{file_name}.json", 'w') as file:
                    json.dump(obj=(header, self.lists, meta), fp=file)
        except TypeError as exc:
            log.exception("Some type error during saving occurred.", exc_info=exc)
            raise
        except PermissionError as exc:
            log.exception(f"Writing permission denied for '{folder}'.", exc_info=exc)
            raise
        else:
            # Indicate the name.
            log.info(f"Saved data to '{folder}/{file_name}'.")
            self.last_save_name = file_name
            return file_name

    def stop_collecting(self) -> None:
        """Stop the data acquisition."""
        log.info("Stopping to collect data.")
        self.trigger_type = TriggerTypes.NONE
        self.unsubscribe_all()
        try:
            self.timer.cancel()
            del self.timer
        except AttributeError:
            pass

    def get_configuration(self) -> dict[str, Any]:
        """Get the currently used configuration as a dictionary."""
        config: dict[str, Any] = {}
        # Trigger
        config['trigger_type'] = self.trigger_type.value
        config['trigger_timeout'] = self.trigger_timeout
        config['trigger_variable'] = self.trigger_variable
        # Value
        vm = ValuingModes.LAST if self.valuing == self.last else ValuingModes.AVERAGE
        config['valuing_mode'] = vm.value
        config['value_repeating'] = self.value_repeating
        # Header and Variables.
        with self.list_lock:
            config['variables'] = list(self.lists.keys())
        config['units'] = self.units
        # config['autoSave'] = self.actionAutoSave.isChecked()
        return config

    def get_last_datapoint(self) -> dict[str, Any]:
        """Read the last datapoint."""
        return self.last_datapoint

    def get_last_save_name(self) -> Union[str, None]:
        """Return the name of the last save."""
        return self.last_save_name

    def get_list_length(self) -> int:
        """Return the length of the lists."""
        with self.list_lock:
            length = len(self.lists[list(self.lists.keys())[0]]) if self.lists else 0
            return length


def main() -> None:
    """Start a datalogger at script execution."""
    parser.description = "Log data."
    parser.add_argument("-d", "--directory",
                        help="set the directory to save the data to")

    gLog = logging.getLogger()  # print all log entries!
    kwargs = parse_command_line_parameters(parser=parser, parser_description="Log data.",
                                           logger=gLog)
    if not gLog.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(StrFormatter)
        gLog.addHandler(handler)

    datalogger = DataLogger(log=gLog, **kwargs)
    datalogger.listen()


if __name__ == "__main__":  # pragma: no cover
    main()



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import pytest
from time import sleep

from pyleco.utils.data_publisher import DataPublisher
from pyleco.utils.listener import Listener
from pyleco.core import PROXY_SENDING_PORT

from pyleco.coordinators.proxy_server import start_proxy, port


pytestmark = pytest.mark.skip("Hangs on teardown in CI")


# Parameters
offset = 100


class ModListener(Listener):
    def __init__(self, name: str, host: str = "localhost", data_port: int = PROXY_SENDING_PORT,
                 **kwargs) -> None:
        super().__init__(name=name, host=host, data_port=data_port, **kwargs)
        self._data: list[dict] = []

    def handle_subscription_data(self, data: dict) -> None:
        self._data.append(data)


@pytest.fixture(scope="module")
def publisher() -> DataPublisher:
    return DataPublisher(full_name="abc", port=port - 2 * offset)


@pytest.fixture(scope="module")
def listener(publisher):
    context = start_proxy(offset=offset)
    listener = ModListener(name="listener", data_port=port - 1 - 2 * offset)
    listener.start_listen()
    listener.communicator.subscribe("")
    sleep(.5)  # due to slow joiner: Allow time for connections.
    yield listener  # type: ignore
    listener.close()
    context.destroy()  # in order to stop the proxy


def test_publishing(publisher: DataPublisher, listener: ModListener):
    publisher.send_data(topic="topic", data="value")
    sleep(.1)
    assert listener._data == [{"topic": "value"}]



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
Zero MQ Proxy server for data exchange.

methods
-------
pub_sub_proxy
    Run a publisher subscriber proxy in the current thread (blocking).
start_proxy
    Start a proxy server, either local or remote, in its own thread.


Execute this module to start a proxy server. If no remote connection given, a
local proxy is created (necessary for remote proxies).
command line arguments:
    -v show all the data passing through the proxy
    -s NAME/IP Subscribe to the local proxy of some other computer
    -p NAME/IP Publish to the local proxy of some other computer

Created on Mon Jun 27 09:57:05 2022 by Benedikt Burger
"""

from __future__ import annotations
import logging
import threading
from typing import Optional

import zmq

if __name__ == "__main__":
    from pyleco.core import PROXY_RECEIVING_PORT
else:
    from ..core import PROXY_RECEIVING_PORT


log = logging.Logger(__name__)

port = PROXY_RECEIVING_PORT


# Technical method to start the proxy server. Use `start_proxy` instead.
def pub_sub_proxy(
    context: zmq.Context,
    captured: bool = False,
    sub: str = "localhost",
    pub: str = "localhost",
    offset: int = 0,
    event: Optional[threading.Event] = None,
) -> None:
    """Run a publisher subscriber proxy in the current thread (blocking)."""
    s: zmq.Socket = context.socket(zmq.XSUB)
    p: zmq.Socket = context.socket(zmq.XPUB)
    _port = port - 2 * offset
    if sub == "localhost" and pub == "localhost":
        log.info(f"Start local proxy server: listening on {_port}, publishing on {_port - 1}.")
        s.bind(f"tcp://*:{_port}")
        p.bind(f"tcp://*:{_port - 1}")
    else:
        log.info(
            f"Start remote proxy server subsribing to {sub}:{_port - 1} and publishing to "
            f"{pub}:{_port}."
        )
        s.connect(f"tcp://{sub}:{port -1 - 2 * offset}")
        p.connect(f"tcp://{pub}:{port - 2 * offset}")

    if captured:
        log.info("Capturing all messages.")
        c: zmq.Socket = context.socket(zmq.PUB)
        c.bind("inproc://capture")
    else:
        c = None  # type: ignore
    if event is not None:
        event.set()
    try:
        zmq.proxy_steerable(p, s, capture=c)
    except zmq.ContextTerminated:
        log.info("Proxy context terminated.")
    except Exception as exc:
        log.exception("Some other exception on proxy happened.", exc)


def start_proxy(
    context: Optional[zmq.Context] = None,
    captured: bool = False,
    sub: str = "localhost",
    pub: str = "localhost",
    offset: int = 0,
) -> zmq.Context:
    """Start a proxy server, either local or remote, in its own thread.

    Examples:

    .. code-block:: python

        # Between software on the local computer, necessary on every computer:
        c = start_proxy()
        # Get the data from a to localhost:
        c = start_proxy(sub="a.domain.com")
        # Send local data to b:
        c = start_proxy(pub="b.domain.com")
        # Send from a to b, can be executed on some third computer:
        c = start_proxy(sub="a.domain.com",
                        pub="b.domain.com")
        # Stop the proxy:
        c.destroy()

    :param context: The zmq context.
    :param bool captured: Print the captured messages.
    :param str sub: Name or IP Address of the server to subscribe to.
    :param str pub: Name or IP Address of the server to publish to.
    :param offset: How many servers (pairs of ports) to offset from the base one.
    :return: The zmq context. To stop, call `context.destroy()`.
    """
    context = context or zmq.Context.instance()
    event = threading.Event()
    thread = threading.Thread(
        target=pub_sub_proxy, args=(context, captured, sub, pub, offset, event)
    )
    thread.daemon = True
    thread.start()
    started = event.wait(1)
    if not started:
        raise TimeoutError("Starting of proxy server failed.")
    log.info("Proxy thread started.")
    return context


def main(
    arguments: Optional[list[str]] = None, stop_event: Optional[threading.Event] = None
) -> None:
    from pyleco.utils.parser import ArgumentParser, parse_command_line_parameters

    parser = ArgumentParser(prog="Proxy server")
    parser.add_argument(
        "-s", "--sub", help="set the host name to subscribe to", default="localhost"
    )
    parser.add_argument("-p", "--pub", help="set the host name to publish to", default="localhost")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase the logging level by one, may be used more than once",
    )
    parser.add_argument(
        "-c",
        "--captured",
        action="store_true",
        default=False,
        help="log all messages sent through the proxy",
    )
    parser.add_argument("-o", "--offset", help="shifting the port numbers.", default=0, type=int)
    kwargs = parse_command_line_parameters(
        parser=parser, logger=log, arguments=arguments, logging_default=logging.INFO
    )

    log.addHandler(logging.StreamHandler())
    if kwargs.get("captured"):
        log.setLevel(logging.DEBUG)
    merely_local = kwargs.get("pub") == "localhost" and kwargs.get("sub") == "localhost"

    if not merely_local:
        log.info(
            f"Remote proxy from {kwargs.get('sub', 'localhost')} "
            f"to {kwargs.get('pub', 'localhost')}."
        )
    else:
        log.info(
            "This data broker manages the data between measurement software, "
            f"which publishes on port {port}, and all the consumers of data "
            f" (DataLogger, Beamprofiler etc.), which subscribe on port {port - 1}."
        )
    context = zmq.Context()
    start_proxy(context=context, **kwargs)
    if merely_local:
        start_proxy(context=context, offset=1)  # for log entries
    reader = context.socket(zmq.SUB)
    reader.connect("inproc://capture")
    reader.subscribe(b"")
    poller = zmq.Poller()
    poller.register(reader, zmq.POLLIN)
    while stop_event is None or not stop_event.is_set():
        if socks := dict(poller.poll(1)):
            if reader in socks:
                received = reader.recv_multipart()
                log.debug(f"Message brokered: {received}")
    context.term()


if __name__ == "__main__":  # pragma: no cover
    main()
